
;---------------------------------------------------------------------------
;# 
;#  The following are sample procs that compute specifications for the
;#  gencyl command of the GL widget.
;#
;#---------------------------------------------------------------------------

(load "TCL")

;proc cylinder { { radius 1 } { height 1 } { nsides 20 } } {
;    # Returns a gencyl specification of a circular cylinder with
;    # the given radius and height
;    return "-close both -polygon $radius $nsides -plot -transl 0 0 $height -plot"
;}

(define-proc-defaults cylinder (( radius 1)( height 1 )(nsides 20))
 `(:close' both :polygon ,radius ,nsides :plot :transl 0 0 ,height :plot)
)
 

;proc cone { { radius 1 } { height 1 } { nsections 10} { nsides 20 } } {
;    # Returns a gencyl specification of a cone with circular base
;    # with the given radius and height. 10 cross sections are
;    # generated in order to yield a smoother shading
;    set cyl "-polygon $radius $nsides "
;    for {set i 0} {$i<=$nsections} {incr i} {
;        lappend cyl -ident -transl 0 0 [expr $height*$i*1.0/$nsections]
;        lappend cyl -scale [expr ($nsections-$i)*0.1] \
;                [expr ($nsections-$i)*0.1] 1 -plot
;    }
;    return $cyl
;}
(define-proc-defaults cone (( radius 1)( height 1 )(nsections 10)(nsides 20))
  (let* ( (cyl 0) )
 ;   # Returns a gencyl specification of a cone with circular base
 ;   # with the given radius and height. 10 cross sections are
 ;   # generated in order to yield a smoother shading
    (set!  cyl `(:polygon ,radius ,nsides))
 
    (do ((i 0 (+ i 1)))
        ((> i nsections) )
        (set!  cyl (append cyl `(:ident :transl 0 0
               ,(*  height i 1.0 (/ nsections)) )))  
        (set!  cyl (append cyl `(:scale ,(* (- nsections i) 0.1)
                ,(* (- nsections i) 0.1) 1 :plot)))
    )
    
   cyl
))

;proc gear { {inner 3} {outer 4} {width 1} {teeth 20} {ncross 4} {torsiondegrees 30} } {
;   lappend cyl -close both -cross
;   set teeth [expr 2*$teeth]
;   for {set tooth 0} {$tooth<$teeth} {incr tooth} {
;      set ang1 [expr 3.141592654*2*$tooth/$teeth]
;      set ang2 [expr 3.141592654*2*($tooth+1)/$teeth]
;      if {$tooth%2==0} { set radius $inner } { set radius $outer } 
;      lappend cyl [expr $radius * cos($ang1)] [expr $radius * sin($ang1)] 0.0 
;      lappend cyl [expr $radius * cos($ang2)] [expr $radius * sin($ang2)] 0.0 
;   }
;   lappend cyl -plot 
;   set dtrans [expr -$width*1.0/$ncross]
;   set drot [expr $torsiondegrees*1.0/$ncross]
;   for {set i 1} { $i < $ncross } {incr i} {
;       lappend cyl -translate 0 0 $dtrans -rotate $drot 0 0 1 -plot
;   }
;   return $cyl
;}
(define-proc-defaults  gear ((inner 3)(outer 4)(width 1)(teeth 20)
       (ncross 4)(torsiondegrees 30))
  (let* ((cyl 0)(ang1 0) (ang2 0)(dtrans 0)(drot 0)(radius 0))
   (set!  cyl '(:close 'both :cross))
   (set!  teeth (*  2 teeth))
   (do ((tooth 0 (+ tooth 1)))
        ((>= tooth teeth))
       (set!  ang1 (*  3.141592654 2 tooth (/ teeth)))
       (set!  ang2 (*  3.141592654 2 (+ tooth 1) (/ teeth)))
      (if (= (modulo tooth 2) 0) ( set!  radius inner)(set!  radius outer))
      (set!  cyl (append cyl 
             `(,(* radius (cos ang1)) ,(* radius ( sin ang1)) 0.0)))
      (set! cyl (append  cyl 
           `(,(* radius (cos ang2)) ,(* radius ( sin ang2))  0.0 )))
    )
   (set! cyl (append cyl '(:plot)))
   (set!  dtrans (*  (- width) 1.0 (/ ncross)))
   (set!  drot (* torsiondegrees 1.0 (/ ncross)))
   (do ((i 1 (+ i 1)))
       ((>=  i ncross))
       (set! cyl(append cyl `(:translate 0 0 ,dtrans :rotate ,drot 0 0 1 :plot)))
   )
   cyl
))

;proc torus { { inner 1 } { outer 10 } { nsections 36 } { nvertices 12 } } {
;    # Returns the gencyl specification of a torus where the cross-section
;    # has radius 'inner' and is rotated along the main axis at a distance
;    # of 'radius'. 'nsections' specifies the number of cross sections
;    # generated. 'nvertices' is the number of vertices per cross section
;    lappend cyl -stitch both -polygon $inner $nvertices 
;    set dang [expr 360.0 / $nsections]
;    for { set i 0 } {$i < 360} {set i [expr $i + $dang]} {
;        lappend cyl -ident -transl -$outer 0 0 -rot $i 0 1 0 -plot
;    }
;    return $cyl
;}
(define-proc-defaults torus ((inner 1)( outer 10)(nsections 36)(nvertices 12))
  (let* ( (cyl ()) (dang 0))
;    # Returns the gencyl specification of a torus where the cross-section
;    # has radius 'inner' and is rotated along the main axis at a distance
;    # of 'radius'. 'nsections' specifies the number of cross sections
;    # generated. 'nvertices' is the number of vertices per cross section
    (set! cyl (append cyl `(:stitch 'both :polygon ,inner ,nvertices)))
    (set! dang (/ 360.0 nsections))
    (do ((i 0 (+ i dang)))
        ((>= i 360)) 
       (set! cyl (append cyl`(:ident :transl ,(- outer) 0 0 :rot ,i 0 1 0 :plot)))
    )
   cyl
))

;proc sphere { { radius 1 } { nsections 18 } { nstripes 20 } } {
;    # Returns the gencyl specification of a sphere with the given radius.
;    # 'nsections' corresponds to the total number of cross-sections and
;    # 'nstripes' corresponds to the number of vertices in each cross-section.
;    set cyl "-polygon $radius $nstripes"
;    set dang [expr 180.0 / $nsections]
;    for {set i -90.0} {$i <= 90.0} {set i [expr $i+$dang]} {
;        set ang [expr $i*3.141592654/180.0]
;        set x [expr $radius * cos($ang)]
;        set y [expr $radius * sin($ang)]
;        lappend cyl -ident -transl 0 0 $y -scale $x $x 1 -plot
;    }
;    return $cyl
;}
(define-proc-defaults  sphere (( radius 1)(nsections 18)(nstripes 20))
    (let* ( (cyl ()) (dang 0)(x 0)(y 0) )
;    # Returns the gencyl specification of a sphere with the given radius.
;    # 'nsections' corresponds to the total number of cross-sections and
;    # 'nstripes' corresponds to the number of vertices in each cross-section.
    (set!  cyl `(:polygon ,radius ,nstripes))
    (set!  dang (/  180.0 nsections))
    (do ((i -90.0 (+ i dang))(ang 0))
        ((> i  90.0)) 
        (set!  ang (* i 3.141592654 (/ 180.0)))
        (set!  x (* radius ( cos ang)))
        (set! y (* radius  (sin ang)))
        (set! cyl (append cyl `(:ident :transl 0 0 ,y :scale ,x ,x 1 :plot)))
    )
   cyl
))

;proc screw { {radius 1} {height 5} {nturns 1} {dsize 0.05} {dang 10} {sides 3} } {
;    # Returns the gencyl specification of a screw, which is a polygon
;    # successively translated scaled and rotated along the z axis.
;    # 'radius' is the initial radius of the cross-section.
;    # 'height' is the total height of the screw.
;    # 'nturns' is the number of complete turns of the shape.
;    # 'dsize' is how much the size of the shape is incremented at each step
;    # 'dang' is how much the shape is turned at each step (degrees)
;    # 'sides' is the number of sides in the cross-section.
;    set cyl "-shade smoothrows -polygon $radius $sides -plot"
;    if $sides==2 { lappend cyl -stitch none }
;    set iz [expr $height*$dang/360.0/$nturns]
;    set n [expr $nturns*360.0/$dang]
;    for {set i 0} {$i < $n} {incr i} {
;        set s [expr $i*$dsize+1]
;        lappend cyl -ident -transl 0 0 [expr $iz*$i] \
;                -rot [expr $i*$dang] 0 0 1 \
;                -scale $s $s 1 -plot
;    }
;    return $cyl
;}
(define-proc-defaults screw ((radius 1)(height 5)(nturns 1)(dsize 0.05)(dang 10)(sides 3))
  (let* ( (cyl ()) (iz 0)(n 0))
;    # Returns the gencyl specification of a screw, which is a polygon
;    # successively translated scaled and rotated along the z axis.
;    # 'radius' is the initial radius of the cross-section.
;    # 'height' is the total height of the screw.
;    # 'nturns' is the number of complete turns of the shape.
;    # 'dsize' is how much the size of the shape is incremented at each step
;    # 'dang' is how much the shape is turned at each step (degrees)
;    # 'sides' is the number of sides in the cross-section.
    (set! cyl `(:shade 'smoothrows :polygon ,radius ,sides :plot))
    (if (= sides 2) (set! cyl(append cyl '(:stitch 'none))))
    (set!  iz (* height dang (/ 360.0)(/ nturns)))
    (set!  n (* nturns 360.0 (/ dang)))
    (do ((i 0 (+ i 1))(s 0))
        ((>= i n))
        (set!  s (+ (*  i  dsize) 1))
        (set! cyl (append cyl `(:ident :transl 0 0 ,(* iz i)
                :rot ,(* i dang)  0 0 1 
                :scale ,s ,s 1 :plot)))
    )
    cyl
))

;proc snail { {r0 1} {d0 0} {alpha 2} {nturns 3} {dang 10} } {
;    # r0 is the initial radius
;    # d0 is the initial distance between the center of the cross-section and
;    #    the axis
;    # alpha is the ratio of size between two full turns of the cross-section
;    # nturns is the total number of turns
;    # dang is the angle increment between two cross sections (degrees)
;    set cyl "-polygon $r0 20"
;    set r $r0
;    set finalang [expr $nturns * 360]
;    for { set i 0 } {$i < $finalang} { incr i $dang } {
;        set k [expr $i/360.0]
;        set r [expr $r0*pow($alpha,$k)]
;        set dst [expr 2*$r0*(1-pow($alpha,$k+1))/(1-$alpha)+$d0-$r]
;        lappend cyl -ident -scale $r $r $r -transl $dst 0 0 \
;                -rot -$i 0 1 0 -plot
;    }
;    return $cyl
;}
(define-proc-defaults snail ((r0 1)(d0 0)(alpha 2)(nturns 3)(dang 10))
   (let* ( (cyl ()) (r 0)(finalang 0) )
;    # r0 is the initial radius
;    # d0 is the initial distance between the center of the cross-section and
;    #    the axis
;    # alpha is the ratio of size between two full turns of the cross-section
;    # nturns is the total number of turns
;    # dang is the angle increment between two cross sections (degrees)
    (set!  cyl `(:polygon ,r0 20))
    (set!  r r0)
    (set! finalang (* nturns  360))
    (do ((i 0 (+ i dang))(k 0)(dst 0))
        ((>= i finalang)) 
        (set!  k (/ i 360.0))
        (set!  r (*  r0  (expt alpha k)))
        (set!  dst (+ (*  2 r0 (- 1 (expt alpha (+ k 1))) (/(- 1 alpha)))
                d0 (- r)))
        (set! cyl (append cyl `(:ident :scale ,r ,r ,r :transl ,dst 0 0
                :rot ,(- i) 0 1 0 :plot)))
    )
   cyl
))

;proc funcrev { {xmin -2} {xmax 2} {nx 20} {nsections 10} 
;               {yfunc {sin($x)+2}} } {
;    # Plots a revolution of the curve yfunc(x) for xmin<=x<=xmax
;    # sampled nx times in the x direction and with nsections sections.
;    # yfunc should be a command that, given a value for x, computes
;    # the corresponding y value
;    set dx [expr ($xmax-$xmin)*1.0/($nx-1)]
;    lappend cyl -stitch ends -cross
;    for {set ix 0} {$ix<$nx} {incr ix} {
;        set x [expr $xmin+$ix*$dx]
;        set y [eval expr $yfunc]
;        lappend cyl $x $y 0
;    }
;    set dang [expr 360.0/$nsections]
;    for {set i 0} {$i<360} {set i [expr $i+$dang]} {
;        lappend cyl -plot -rot $dang 1 0 0 
;    }
;    return $cyl
;}
(define-proc-defaults funcrev ((xmin -2)(xmax 2)(nx 20)(nsections 10) 
               (yfunc (lambda (x) (+ (sin x) 2) ))  )
    (let* ( (cyl()) (dx 0) (dang 0) )
;    # Plots a revolution of the curve yfunc(x) for xmin<=x<=xmax
;    # sampled nx times in the x direction and with nsections sections.
;    # yfunc should be a command that, given a value for x, computes
;    # the corresponding y value
    (set! dx (*  (- xmax xmin) 1.0(/ (- nx 1))))
    (set! cyl (append cyl '(:stitch 'ends :cross)))
    (do ((ix 0 (+ ix 1))(x 0)(y 0))
        ((>= ix nx))
        (set!  x (+ xmin (* ix dx)))
        (set!  y (yfunc x))
        (set! cyl (append cyl `(,x ,y 0)))
    )
    (set!  dang (/  360.0 nsections))
    (do ((i 0 (+ i dang)))
        ((>= i 360))
        (set! cyl (append cyl `(:plot :rot ,dang 1 0 0)))
    )
  cyl
))

;proc funcmesh {{xmin -5} {xmax 5} {ymin -5} {ymax 5} {nx 20} {ny 20} 
;               {zfunc {2*cos(sqrt($x*$x+$y*$y))}} } {
;    # Plots a mesh of zfunc (x,y) for points xmin<=x<=xmax and 
;    # ymin<=y<=ymax sampled nx times in the x direction and ny times
;    # in the z direction. zfunc should be a command that takes two
;    # numeric variables (x,y) and computes the corresponding z value
;    # as the result
 ;   set dx [expr ($xmax-$xmin)*1.0/($nx-1)]
 ;   set dy [expr ($ymax-$ymin)*1.0/($ny-1)]
 ;   set cyl "-stitch none"
;    for {set ix 0} {$ix<$nx} {incr ix} {
;        set x [expr $xmin+$ix*$dx]
;        lappend cyl -ident -transl 0 0 $x -cross
;        for {set iy 0} {$iy<$ny} {incr iy} {
;            set y [expr $ymin+$iy*$dy]
;            set z [eval expr $zfunc]
;            lappend cyl $y $z 0
;        }
;        lappend cyl -plot
;    }
;    return $cyl
;}
(define-proc-defaults funcmesh ((xmin -5)(xmax 5)(ymin -5)(ymax 5)(nx 20)(ny 20)
            (zfunc (lambda (x y)(* 2  (cos (sqrt (+ (* x x)(* y y)))))))
            )
  (let* ((cyl ())  (dx 0)(dy 0) )
;    # Plots a mesh of zfunc (x,y) for points xmin<=x<=xmax and
;    # ymin<=y<=ymax sampled nx times in the x direction and ny times
;    # in the z direction. zfunc should be a command that takes two
;    # numeric variables (x,y) and computes the corresponding z value
;    # as the result
    (set!  dx (* (- xmax xmin) 1.0(/(- nx 1))))
    (set!  dy (* (- ymax ymin) 1.0(/(- ny 1))))
    (set! cyl '(:stitch 'none))
    (do ((ix 0(+ ix 1)) (x 0)(y 0)(z 0))
        ((>= ix nx))
        (set!  x (+ xmin (* ix dx)))
        (set! cyl (append cyl `(:ident :transl 0 0 ,x :cross)))
        (do ((iy 0(+ iy 1)))
            ((>= iy ny)) 
           ( set!  y (+ ymin (* iy dy)))
           ( set!  z (zfunc x y)) 
           (set! cyl (append cyl (list y z 0)))
        )
       (set! cyl (append cyl '(:plot)))
    )
   cyl
))

#|
#---------------------------------------------------------------------------
#
# General utility procs for the GL widget
#
#---------------------------------------------------------------------------

proc centercamera { min max varname } {
    # Computes a default camera that centers an object given its bounding box
    # 'min' and 'max' are lists of the form "x y z".
    # As side effects, elements 'center', 'eye', 'size', 'znear' and 'zfar'
    # of global array  'varname' are set
    upvar #0 $varname camera
    scan $min "%f %f %f" xmin ymin zmin
    scan $max "%f %f %f" xmax ymax zmax
    
    set camera(center) "[expr ($xmax+$xmin)/2] [expr ($ymax+$ymin)/2] \
            [expr ($zmax+$zmin)/2]"
    set total 0
    foreach c {x y z} {
        set side [expr ([set ${c}max]-[set ${c}min])/2]
        set total [expr $total+$side*$side]
    }
    set camera(size) [expr sqrt($total)]
    set camera(eye) "[expr ($xmax+$xmin)/2] [expr ($ymax+$ymin)/2] \
            [expr ($zmax+$zmin)/2+2*$camera(size)]"
    set camera(field) 60
    set camera(znear) $camera(size)
    set camera(zfar) [expr 3*$camera(size)]
}
|#
#|
(define-syntax  scanf
   (syntax-rules ()
      ( (scanf :do-rest inp)  ())
      ( (scanf :do-rest inp var1 var2 ...)  
             (begin (set! var1  (read inp)) (scanf :do-rest inp var2 ...)))
      ( (scanf  str var1 var2 ...)
         (let* ((in (open-input-string str)))
            (set! var1 (read in))
            (scanf  :do-rest in var2 ...)
            (close-input-port in)
         )
           )
   )
)
|#
   
(define (centercamera  min max camera)
  (let* ((xmin 0)(ymin 0)(zmin 0)(xmax 0)(ymax 0)(zmax 0) (total 0)(side 0))
      (scan-list min xmin ymin zmin)  
      (scan-list  max xmax ymax zmax)  
       (set-prop  camera  'center (list (/ (+ xmax xmin) 2) (/ (+ ymax ymin) 2)
               (/ (+ zmax zmin) 2)))
      (set!  total 0)
      (set!  side (/  (- xmax xmin) 2))
      (set!  total (+ total (* side side)))
       (set!  side (/  (- ymax ymin) 2))
      (set!  total (+ total (* side side)))
        (set!  side (/  (- zmax zmin) 2))
      (set!  total (+ total (* side side)))
      (set-prop  camera 'size  (sqrt  total))
       (set-prop  camera 'eye (list (/ (+ xmax xmin) 2)(/ (+ ymax ymin)2)
                   (+  (/ (+ zmax zmin) 2) (* 2 (get-prop camera 'size)))))
       (set-prop  camera 'field 60)
       (set-prop  camera 'znear (get-prop camera 'size))
       (set-prop  camera 'zfar   (* 3 (get-prop camera  'size)))
)) 
      
#|
proc redraw {} {
    # sets up widget 'gl1' for redraw. The object to be drawn should
    # have been set up previously and the corresponding display list should
    # be in global variable 'dl'. Other global variables are used
    # to describe the scene:
    # 'xrot' and 'yrot' are angles in degrees for rotating the object
    # 'pmode' should be a valid constant for option -polygonmode
    # 'camera' is a global array containing the various options for
    #     the view command

    global xrot yrot dl pmode camera shade
    scan $camera(center) "%f %f %f" xc yc zc
    if {$shade=="flat"} {
        set shademodel flat
    } else {
        set shademodel smooth
    }
    eval ".gl1 main \
            -clear depthbuffer colorbuffer \
            -polygonmode front $pmode \
            -shademodel $shademodel \
            -matrixmode projection \
            -loadidentity \
            -perspective 60 1 $camera(znear) $camera(zfar)\
            -matrixmode modelview \
            -loadidentity \
            -lookat $camera(eye) $camera(center) 0 1 0 \
            -translate $xc $yc $zc \
            -rotate $yrot 1 0 0 \
            -rotate $xrot 0 1 0 \
            -translate [expr -$xc] [expr -$yc] [expr -$zc]\
            -call $dl"
}
|#

(define   (redraw)
  (let* ((xc 0)(yc 0)(zc 0))
     (scan-list (get-prop 'camera  'center) xc yc zc)
     (if (equal? shade "flat")
        (set!  shademodel 'flat)
        (set!  shademodel 'smooth)
     )
     (eval `(.gl1 'main 
            :clear 'depthbuffer 'colorbuffer
            :polygonmode 'front pmode 
            :shademodel shademodel 
            :matrixmode 'projection 
            :loadidentity 
            :perspective 60 1 (get-prop 'camera 'znear) (get-prop 'camera 'zfar)
            :matrixmode 'modelview 
            :loadidentity 
            :lookat ,@(get-prop 'camera 'eye) ,@(get-prop 'camera 'center) 0 1 0 
            :translate ,xc ,yc ,zc
            :rotate yrot 1 0 0 
            :rotate xrot 0 1 0 
            :translate (-  ,xc)(- ,yc)( - ,zc)
            :call dl 
      ))
))
#|
proc makemodel args {
    # Calls the gencyl command of widget 'gl1' with the result of
    # the command line expressed in 'args'. Sets up a default camera
    # and deletes previous display lists that may have been defined earlier
    global dl camera shade adaptive adaptivethreshold
    if {$adaptive} {
        set adaptiveoption "-adaptive $adaptivethreshold"
    } else {
        set adaptiveoption ""
    }
    set info [eval .gl1 gencyl -texgen 0 1 0 1 $adaptiveoption -shade $shade [$args]]
    centercamera [getpropval $info min] [getpropval $info max] camera
    catch {.gl1 deletelist $dl}
    set dl [getpropval $info displaylist]
    redraw
}
|#
;;;(("displaylist" 5) ("min" -1 -1 0) ("max" 1 1 1))
(define (getpropval info-list prop)
  (let* ((res ()))
     (if (symbol? prop) (set! prop (symbol->string prop)))
     (set! res (assoc prop info-list))
     (if res
          (begin
               (set! res (cdr res))  ;;;; return a list
               (if (eqv? (length res) 1)
                    (car res) ;;;; for displaylist
                    res    
               )
           )
          ()
     )
  )
)

(define (makemodel   args)
  (let* ((adaptiveoption #f) (opt 0))
   (if (string? args) (set! args (string->symbol args)))
   (format #t "args ~s ~%" args)  ;;;; for debug
   (if (not (eqv? adaptive 0)) ;;;if adaptive 
        (set!  adaptiveoption (list :adaptive  adaptivethreshold))
        (set!  adaptiveoption ())
   )
   (format #t "adaptive ~s ~%" adaptiveoption)  ;;;; for debug
   ;;;(set! opt ((function args)))
   (set! opt (funcall args))
   (format #t "opt ~s ~%" opt)  ;;;; for debug
   (set!  info (eval `(.gl1 'gencyl :texgen 0 1 0 1  ,@adaptiveoption 
             :shade ,shade ,@opt)))
   (format #t "makemodel info~s ~%" info)  ;;;; for debug
  ;;;(("displaylist" 5) ("min" -1 -1 0) ("max" 1 1 1))
  ;;;; ????? getpropval
    (centercamera (getpropval info 'min) (getpropval info 'max) 'camera)
     (catch (.gl1 'deletelist dl))
    (set!  dl (getpropval info 'displaylist))
    (redraw)
))

#|
proc createcheckerimg {} {
    # Creates a simple texture image called img
    image create photo tmp -width 2 -height 2
    image create photo img -width 64 -height 64
    tmp put {{#ffffff #ff0000} {#000000 #ffffff}} 
    img copy tmp -from 0 0 2 2 -zoom 4 4 -to 0 0 64 64
    rename tmp {}
}
|#
(define ( createcheckerimg )
   ;;; # Creates a simple texture image called img
   (image 'create 'photo 'tmp :width 2 :height 2)
   (image  'create 'photo 'img :width 64 :height 64)
   ;;(tmp 'put '((#xffffff #xff0000)(#x000000 #xffffff)))
   (tmp 'put '(("#ffffff" "#ff0000")("#000000" "#ffffff")))
    (img 'copy 'tmp :from 0 0 2 2 :zoom 4 4 :to 0 0 64 64)
    ;;;rename tmp {}
)
#|
#---------------------------------------------------------------------------
#
# Procs to set up a user interface to examine and modify the default
# arguments of a procedure.
# 
#---------------------------------------------------------------------------

proc updatedefarg { procname } {
    global body arglist defarg
    foreach arg $arglist {
        lappend argspec [list $arg $defarg($arg)]
    }
    eval [list proc $procname $argspec $body]
}
|#
;;;  ?????
(define (updatedefarg  procname )
   ;;; global body arglist defarg
   (let* ((argspec ()) )
      (foreach arg arglist 
         ;;(push (list arg (get-prop 'defarg arg)) argspec)
         (push (list arg (symbol-value arg)) argspec)
       )
    ;;; (format #t "updatefarg argspec ~s" argspec)
    (set-prop procname 'default argspec)
  )
)

#|
proc setupdefargs { path procname updatecmd } {
    frame $path 
    global arglist defarg body
    set arglist [info args $procname]
    set body [info body $procname]
    foreach arg $arglist {
        if ![info default $procname $arg tmp] {
            set defarg($arg) {}
        } else {
            set defarg($arg) $tmp
        }
        pack [frame $path.$arg] -side top -expand yes -fill x        
        label $path.$arg.lab -text $arg
        entry $path.$arg.ent -textvar defarg($arg) 
        bind $path.$arg.ent <Key-Return> "updatedefarg $procname; eval $updatecmd"
        pack $path.$arg.lab -side left
        pack $path.$arg.ent -side right
    }
    return $path
}
|#
;;;  ????
(define (setupdefargs  path procname updatecmd )
  (let* ((temp 0))
    (frame path)
    ;;;global arglist defarg body
    ;;;(if (string? procname) (set! procname (string->symbol procname)))
    (set!  arglist (proc-info  args procname))
    ;;;(set!  body (proc-info body procname))
    (foreach arg arglist 
        (if (not (proc-info default procname arg temp))
            (set-prop  'defarg  arg ())
            (set-prop   'defarg arg temp)
        )
        ;;;(format #t "arg ~s temp ~s ~%" arg temp)
        (pack (frame (tk-name path #\. arg)) :side 'top :expand 'yes :fill 'x)
        (label (tk-name path #\. arg ".lab") :text arg)
        (entry (tk-name path #\. arg ".ent")  :textvar  arg) ;;;original can't use 
         (symbol-value-set arg temp)
        (bind (tk-name path #\. arg ".ent") "<Key-Return>" 
               (lambda() (updatedefarg procname)(eval updatecmd)))
        (pack (tk-name path #\. arg ".lab")  :side 'left)
        (pack (tk-name path #\. arg ".ent")  :side 'right)
    )
   path
))

#|
#---------------------------------------------------------------------------
#
# Main program
#
#---------------------------------------------------------------------------
|#
;;; source "setutils.tcl"     ;;;; ????????????
(createcheckerimg)

(OGLwin '.gl1 :height 400 :width 400)
(pack .gl1 :expand 'yes :fill 'both :side 'left)
(define xrot 0) 
(define yrot 0)
(define pmode 'fill)
(define edit #f)  ;;;????   need ?????   not exists in original .tcl  
(define mname 0)  ;;;????   need ?????   not exists in original .tcl  
(define arglist ())  ;;;????   need ?????   not exists in original .tcl  
(define body  ())  ;;;????   need ?????   not exists in original .tcl  
(define x 0)  ;;;????   need ?????   not exists in original .tcl  
(define y 0)  ;;;????   need ?????   not exists in original .tcl  
(define dl 0)  ;;;????   need ?????   not exists in original .tcl  
(define shademodel 0)  ;;;????   need ?????   not exists in original .tcl  

(bind .gl1 "<Button-1>"  (lambda (x y) (symbol-value-set  'x x)(symbol-value-set  'y y)))

(bind .gl1 "<B1-Motion>" 
   (lambda (x y) 
    (incf xrot (-  x  (symbol-value 'x))) 
    (incf yrot (-  y  (symbol-value 'y))) 
    (symbol-value-set  'x x)
    (symbol-value-set  'y y)
    (redraw)
))

(bind .gl1 "<Button-2>" 
    (lambda() 
       (if (equal? pmode "fill") 
             (set!  pmode 'line)
             (set!  pmode 'fill)
       ) 
     ( redraw) )
)

(bind .gl1 "<Enter>" (lambda() (focus .gl1 )
   (.gl1 'eval   :light 'light0 'ambient 1 1 1 1 
            :light 'light0 'diffuse 1 1 1 1 
            :light 'light0 'position 1 1 1 0 
            :light 'light0 'specular 1 1 1 1 
            :enable 'lighting :enable 'light0 
            :enable 'depthtest 
            :teximage2d 0 0 'img
	         :enable 'depthtest 
	         :texparameter 'texture2d 'texturewraps 'repeat 
	         :texparameter 'texture2d 'texturewrapt 'repeat 
	         :texparameter 'texture2d 'texturemagfilter 'nearest 
	         :texparameter 'texture2d 'textureminfilter 'nearest 
	         :texenv 'textureenv 'textureenvmode 'modulate
	         :material 'front 'ambient 0.4 0.4 0.4 
            :material 'front 'diffuse 0.5 0.5 0.5 
            :material 'front 'specular 0.5 0.5 0.5 
            :material 'front 'shininess 50 
            :material 'back 'ambient 0.5 0 0
            :material 'back 'diffuse 0.5 0 0 
            :material 'back 'specular 0 0 0 
            :material 'back 'shininess 50 
     )
  )
)
;;;;#            -lightmodel lightmodeltwoside 1


(pack (frame '.mainframe) :side 'left :fill 'y)
(foreach f '(cmdframe shadeframe modelframe)
    (frame (tk-name #\. f) :bd 2 :relief 'raised)
    (pack (tk-name #\. f) :fill 'both :in  .mainframe :expand 'yes)
)


(button '.quit :text "Quit" :command '(exit))

(define (editcmd) 
      (catch (destroy .editpopup.w))
    (if (not (eqv? edit 0)) 
        (if (not (winfo 'exists '.editpopup))(begin ( toplevel '.editpopup )
          ( pack (setupdefargs '.editpopup.w (tk-name mname)
               (list 'makemodel  (tk-name mname))) :side 'right)))
       ;;else
       (catch (destroy .editpopup))
    )
)

(checkbutton '.edit :text "Edit" :var 'edit :command '(editcmd))
(checkbutton '.texture :text "texture" :var 'texture :command 
  '(begin
     (if (not (eqv? texture 0)) (.gl1 'eval :enable 'texture2d)
            (.gl1 'eval :disable 'texture2d))
     (.gl1 'redraw)
   )
)

(pack .quit .edit .texture :in .cmdframe :side 'left :fill 'x)

(define  shade 'smooth)
(define  adaptive #f)
(define  adaptivethreshold 30)
(define  texture 0)
(radiobutton '.flat :text "Flat Shading" :value 'flat :var 'shade
        :command '(makemodel mname)  )

(radiobutton '.smooth :text "Smooth Shading" :value 'smooth :var 'shade 
        :command  '( makemodel mname ))

(radiobutton '.smoothrows :text "Smooth Shading (rows)" 
        :value 'smoothrows :var 'shade 
        :command '(makemodel mname ))

(radiobutton '.smoothcols :text "Smooth Shading (columns)" 
        :value 'smoothcols :var 'shade 
        :command '(makemodel mname))

(checkbutton '.adaptive :text "Adaptive Shading" :var 'adaptive 
        :command '(makemodel mname ))

(entry '.adaptivethreshold :textvar 'adaptivethreshold )

(bind .adaptivethreshold "<Key-Return>"  (lambda()(makemodel mname )))

(pack .flat .smooth .smoothrows .smoothcols 
        :in .shadeframe :expand 'yes :fill 'y :anchor 'w)

(pack .adaptive .adaptivethreshold 
        :in .shadeframe :side 'left :expand 'yes :fill 'y) 

(set! mname 'cylinder)
(trace makemodel)
(makemodel mname)

(foreach model '(cylinder sphere cone torus gear screw snail funcrev funcmesh )
    (radiobutton (tk-name ".modelframe." model) :text model 
            :var 'mname :value model 
            :command '(begin (makemodel mname)(set! edit 0) ( editcmd)))  
    (pack (tk-name ".modelframe." model)  :expand 'yes :fill 'y :anchor 'w)
)

