;;package require Tkogl
(load "TCL")
(load "tclglob")
(load "tkoptionmenu")
;-----------------------------------------------------------
;
; The 5 following (define (s init the vertices and face lists 
; describing the 5 platonic polyhedra
;
;-----------------------------------------------------------

(define ( init_tetrahedron )
    ;;global poly
  (let* ( ( a (/ 1.0 (sqrt 3.0))) (-a (- a)))
         (setf ($  'poly 'tetrahedron 'vtx)  (list 
	(list a a a) (list a -a -a) 
        (list -a a -a) (list -a -a a) ) )
    (setf  ($  'poly 'tetrahedron 'face) '(
        (0 3 1) (2 0 1) (3 0 2) (1 3 2)
    ) )
))

(define ( init_hexahedron )
    ;;global poly
 (let* ( ( a (/ 1.0 (sqrt 3.0)))  (-a (- a)))
    (setf  ($ 'poly 'hexahedron 'vtx) (list 
	(list a a a) (list a a -a) 
	(list -a a -a) (list -a a a)
	(list -a -a a) (list -a -a -a) 
	(list a -a -a) (list a -a a) ))
    (setf ($  'poly 'hexahedron 'face)  '(
        (0 1 2 3) (3 2 5 4) (4 5 6 7)
        (7 6 1 0)  (1 6 5 2) (0 3 4 7)
    ))
))

(define ( init_octahedron )
    ;;global poly
    (setf ($ ' poly 'octahedron 'vtx) '(
	(1 0 0) (0 1 0) (-1 0 0) (0 -1 0) (0 0 1) (0 0 -1) 
    ))
    (setf  ($ 'poly 'octahedron 'face) '(
        (0 1 4) (1 2 4) (2 3 4) (3 0 4)
        (1 0 5) (2 1 5) (3 2 5) (0 3 5)
    )) 
)

(define ( init_icosahedron )
    ;;global poly
 (let* (   ( X 0.525731112119133606)
    ( Z 0.850650808352039932)  (-X (- X)) (-Z (- Z)))
    (setf ($ ' poly 'icosahedron 'vtx)
	(list (list  -X 0.0 Z) (list X 0.0 Z)
	     (list -X 0.0 -Z) (list X 0.0 -Z) 
	     (list 0.0 Z X) (list 0.0 Z -X) 
	     (list 0.0 -Z X) (list 0.0 -Z -X) 
	     (list Z X 0.0) (list -Z X 0.0) 
	     (list Z -X 0.0) (list -Z -X 0.0) ))
    (setf  ($  'poly 'icosahedron 'face) '( 
	(4 0 1)    (9 0 4)    (5 9 4)    (5 4 8)
	(8 4 1)    (10 8 1)   (3 8 10)    (3 5 8)
	(2 5 3)    (7 2 3)    (10 7 3)    (6 7 10)
	(11 7 6)    (0 11 6)    (1 0 6)    (1 6 10)
	(0 9 11)    (11 9 2)    (2 9 5)    (2 7 11)
    ))
))

(define ( init_dodecahedron )
    ;;global poly
   (let* ( ( alpha (/ (sqrt(/ 2.0(+ 3.0  (sqrt 5.0)))) (sqrt 3.0)))
           (beta (+ 1.0 (sqrt (/ 6.0 (+ 3.0 (sqrt 5.0)))) -2.0 
	        (* 2.0 (sqrt (/ 2.0 (+ 3.0 (sqrt 5.0)))) (/ (sqrt 3.0)))))
          (gamma (/ 1.0 (sqrt 3.0)))
          (-alpha (- alpha)) (-beta (- beta))(-gamma (- gamma)))
    (setf ($ 'poly  'dodecahedron 'vtx) (list 
	(list -alpha  0  beta) 
	(list alpha  0  beta ) 
	(list -gamma  -gamma  -gamma ) 
	(list -gamma  -gamma  gamma ) 
	(list -gamma  gamma  -gamma ) 
	(list -gamma  gamma  gamma ) 
	(list gamma  -gamma  -gamma ) 
	(list gamma  -gamma  gamma ) 
	(list gamma  gamma  -gamma ) 
	(list gamma  gamma  gamma ) 
	(list beta  alpha  0 ) 
	(list beta  -alpha 0 ) 
	(list -beta  alpha  0 ) 
	(list -beta  -alpha 0 ) 
	(list -alpha  0  -beta ) 
	(list alpha  0  -beta ) 
	(list 0  beta  alpha ) 
	(list 0  beta  -alpha ) 
	(list 0  -beta  alpha ) 
	(list 0  -beta  -alpha)))
    (setf ($ 'poly 'dodecahedron 'face) '(
	(0 1 9 16 5) 	(1 0 3 18 7) 
	(1 7 11 10 9)	(11 7 18 19 6) 
	(8 17 16 9 10)	(2 14 15 6 19) 
	(2 13 12 4 14)	(2 19 18 3 13) 
	(3 0 5 12 13)   (6 15 8 10 11) 
	(4 17 8 15 14) 	(4 12 5 16 17) 
    )) 
))

;-----------------------------------------------------------------
;
;  make_stellated creates a stellated polyhedron out of a regular
;  polyhedron by inserting a new vertex in the center of the faces
;  and joining it to the face vertices by triangles. ratio is the
;  relative distance between the inserted vertex and the origin: 
;  -1 brings it close to the origin, 0 corresponds to a vertex lying
;  on the respective face, while 1+ brings it farther away.
;
;----------------------------------------------------------------

(define ( make_stellated  pname newpname ratio ) 
    ;;global poly
   (let* ( ( nface (llength ($ 'poly pname 'face)))
           (nvtx   (llength ($ 'poly  pname 'vtx))) (v 0)(nvtxface 0) )
    (setf ($ 'poly newpname 'vtx) ($ 'poly  pname  'vtx))
    ;;; need initialize,  since in TCL   ""=()=  0  !!!
       (do (( k 0 (+ k 1)))
            ( (>= k 3))
            (setf ($ 'm k) 0 )
        )
    ;;;;

    (do ((i 0 (+ i 1)))
        ( (>= i nface) )
	(set! v (lindex ($ 'poly pname 'face) i))
        (set! nvtxface (llength v))
        (do ((k 0 (+ k 1))) (>= k 3)(setf ($ 'm k) 0.0 ))
        (do ((j 0 (+ j 1)) (vv 0))
            ( (>= j nvtxface)) 
            ;;;(format #t "stellated v ~s  lindex v j  ~s~%" v (lindex v j))
            (set! vv (lindex ($ 'poly pname 'vtx) (lindex v j)))
            ;;;(format #t "stellated vv ~s ~%" vv)
            (do (( k 0 (+ k 1)))
                ( (>= k 3)) 
		(setf ($ 'm k) (+ ($ 'm k)  (lindex vv k)))
            )
        )
        (do (( k 0 (+ k 1)))
            ( (>= k 3))
            (setf ($ 'm k) (* ($ 'm k) (/ nvtxface) (+ 1.0 ratio)))
        )
        (lappend ($ 'poly newpname 'vtx) (list  ($ 'm 0) ($ 'm 1) ($ 'm 2)))
        (do ((j 0(+ j 1))(jj 0))
            ( (>= j  nvtxface))
            (set! jj (modulo  (+ j 1) nvtxface))
            (lappend ($ 'poly newpname 'face) (list 
                (lindex v j) (lindex v jj) nvtx) )
        )
        (incf nvtx)
    )
))

;----------------------------------------------------------------------
;
;  faces_of_vertex returns a list where each element is a list of the
;  form ( f v0 v1 ) where:
;     f is the number of a face of poly $pname adjacent to $ivtx
;     v0 is the circular predecessor of $ivtx
;     v1 is the circular successor of $ivtx
;
;-------------------------------------------------

(define (faces_of_vertex  pname ivtx ) 
    ;;global poly
 (let* ((iface 0) (i 0)(tmp ())(head 0) (elist ()) (nv 0)(i0 0)(i1 0)(edge 0))
    (foreach f ($ 'poly pname 'face) 
        ;;(format #t "faces f ~s ~%" f)
        (set! i (lsearch  f ivtx) )  ;;; -exact 
        (if  i ;;;;(!= i -1)
           (begin
             (set! nv (llength f))
             (set! i0 (modulo (+ i -1 nv) nv))
             (set! i1 (%  (+ i 1) nv))
             (lappend tmp (list iface (lindex f i0) (lindex f i1)))
           )
        )
        (incf iface)
    )
       ;;; (format #t "faces tmp  ~s ~%" tmp)
    (set! head (lindex tmp 0))
    (set! tmp (lrange tmp 1 'end))
    (lappend elist head)
    (while (pair? tmp )
        (set! edge (lindex head 2))
        ;;;(format #t "edge ~s tmp ~s ~%" edge tmp)
        (set! i (lsearch-test  -glob tmp (tk-string "* "  edge " *")))
        ;;(if (= i -1)  (error 7777 "error ordering faces"))
        (if (not  i )  (error 7777 "error ordering faces"))
        (set! head (lindex tmp i))
        (set! tmp (lreplace tmp i i))
        (lappend elist head)
    )
   elist
))

;----------------------------------------------------------------------
;
;  make_truncated makes a truncated polyhedron out of a regular 
;  polyhedron by creating a new face for every vertex of the original
;  polyhedron. ratio is the relative length of the edge which is
;  cut from each original edge.
;
;----------------------------------------------------------------------

(define ( make_truncated  pname newpname ratio ) 
    ;;global poly
  (let* (( nvtx (llength ($ 'poly pname 'vtx)))
         (ratio1 (-  1.0  ratio))
         (newnvtx 0)  (elist ()) (prevnvtx 0) (newface ())  
         (ivtx1 0)(iface 0) (v0 0) (v1 0) (newv ())  )
    (do (( ivtx 0 (+ ivtx 1)))
        (( >= ivtx nvtx))
        ;;(format #t "faces in ivtx ~s ~%" ivtx)
        (set! elist (faces_of_vertex pname ivtx))
       ;;; (format #t "faces out~%")
        (set! prevnvtx (+ newnvtx (llength elist) -1))
        (set! v0 (lindex ($ 'poly pname 'vtx) ivtx))
        (set! newface ())
        (foreach e elist 
            (set! ivtx1 (lindex e 2))
            (set! iface (lindex e 0))
            (set! v1 (lindex ($ 'poly pname 'vtx) ivtx1))
            (set! newv ())
            (do (( c 0 (+ c 1)))
                ((>= c 3))            
                (lappend newv (+ (* (lindex v0 c) ratio1) 
                                  (*  (lindex v1 c) ratio)))
            )
            (lappend ($ 'poly newpname 'vtx) newv)
            (set! newface (linsert newface 0 newnvtx))
            (setf ($ 'vtxsubst ivtx iface) (list prevnvtx newnvtx))
            (set! prevnvtx newnvtx)
            (incf newnvtx)
        )
        (lappend ($ 'poly newpname 'face) newface)
    )
    (set! iface 0)
    (foreach f ($ 'poly pname 'face) 
        (set! newface ())
        (foreach ivtx  f 
            (lappend newface ($ 'vtxsubst ivtx iface))
        )
        (lappend  ($ 'poly newpname 'face) (join newface))
        (incf iface)
    )      
))

;----------------------------------------------------------------------
;
; plot_poly plots all faces of polyhedron $pname in display lists
; whose numbers are stored in global variable poly($pname,dlists) 
;
;---------------------------------------------------------------------- 

(define ( plot_poly  path pname ) 
    ;;global poly
 (if (tcl-info exists ($ 'poly pname 'dlists))
          () ;;; do  nothing
   (let* ((nface 0) (dl 0) (ll ()) (v 0) (nvtxface 0) (vv 0) (nsize 0))    
     (set! nface (llength ($ 'poly pname 'face)))
    (do ((i 0 (+ i 1)))
        ( ( >= i nface))
	(set! v (lindex ($ 'poly pname 'face) i))
	(set! dl ((function path) 'newlist))
	(lappend ($ 'poly pname 'dlists) dl)
        (set! nvtxface (llength v))
	(do ((k 0 (+ k 1)))
            ( (>= k 3)) 
            (setf ($ 'n k) 0.0 )
        )
	((function path)  'eval :newlist dl 'compile :begin 'polygon )
	(set! ll ())
	(do ((j 0 (+ j 1)))
            ( (>= j nvtxface))
	    (set! vv (lindex ($ 'poly pname 'vtx) (lindex v j)))
	    ;;;;(lappend ll (list :vertex vv))
	    (set!  ll (append ll (cons :vertex vv)))
	    (do (( k 0 (+ k 1)))
                ( (>= k 3))  
		(setf ($ ' n k) (+ ($ 'n k) (lindex vv k)))
	    )
	)
        (set! nsize (sqrt (+ (* ($ 'n 0) ($ 'n 0)) (* ($ 'n 1) ($ 'n 1))
                             (* ($ 'n 2) ($ 'n 2)))))
	((function path) ' eval :normal (/ ($ 'n 0) nsize) 
	    (/ ($ 'n 1)  nsize) 
	    (/ ($ 'n 2)  nsize))
	;;;((function path) 'eval ll :end :endlist)
               ;;;; this cause error because tcl-invoke pass ll to tcl
               ;;;; directly as ONE parameter!!!
         (set! ll (append ll (list :end :endlist)))
         (set! ll (cons 'eval ll))
         (apply (function path) ll)
    )
  )
 )
)

;----------------------------------------------------------------------
;
;  rotation_bindings creates window bindings that perform rotation 
;  transformations corresponding to mouse drag actions
;
;----------------------------------------------------------------------
(define xrot 0)
(define yrot 0)
(define x 0)
(define y 0)
(define ( rotation_bindings  path ) 
    ;;global xrot yrot
    (set! xrot 0 )
    (set! yrot 0)
    (bind path "<Button-1>"  (lambda (x y) (symbol-value-set 'x x)(symbol-value-set 'y y )))
    (bind path "<B1-Motion>" ;;;subst -nocommands  
	(lambda (x y) (incf xrot (-  x (symbol-value 'x) ) )
	(incf yrot (- y (symbol-value 'y))) 
	(symbol-value-set 'x x)
	(symbol-value-set   'y y)
	((function path) 'eval :matrixmode 'modelview 
            :popmatrix 
	    :pushmatrix 
            :rotate yrot 1 0 0
            :rotate xrot 0 1 0 
         )
	((function path) 'redraw)
        )
     )
)

;----------------------------------------------------------------------
;
;  init_polybuild sets up the 'build' and 'require' data structures
;  which describe how a given polyhedron name is to be built
;
;----------------------------------------------------------------------

(define ( init_polybuild )
    ;;global build require
    (foreach p '(tetrahedron hexahedron octahedron dodecahedron
               icosahedron) 
        (setf ($ 'build p) (list (tk-name "init_" p))) ;;; to be eval
        (setf ($ 'require p) (list p))     ;;;; must list , used foreach  see polybuild
        (setf ($ 'build  'stellated p)  `( 
	    make_stellated  ',p  '(stellated ,p) 1) )
        (setf ($ 'require 'stellated p) (list p (list 'stellated p)))
	(setf ($ 'build  'sunken p)  `( 
	    make_stellated ',p  '(sunken ,p) -0.5))
	(setf ($ 'require  'sunken p) (list p (list 'sunken p)))
        (setf ($ 'build  'half-truncated p)  `( 
	    make_truncated ',p  '(half-truncated ,p) 0.333))
        (setf ($ 'require ' half-truncated p) (list p 
            (list 'half-truncated p)))
        (setf ($ 'build  'truncated p)  `( 
	    make_truncated ',p  '(truncated ,p) 0.5))
        (setf ($  'require  'truncated p) (list p 
            (list 'truncated p)))
        ;;(set! require((list truncated $p)) (list $p 
        ;;    (list truncated $p)))
    )     
)

;----------------------------------------------------------------------
;
;  polybuild builds a display list for a polyhedron given its name
;
;----------------------------------------------------------------------

(define ( polybuild  path type variation dlist mode ) 
    ;;global poly build require bname
    (set! type (tk-name type))  ;;;;  type is string 
    (set! variation (tk-name variation))  ;;;;  variation is string 
    (set! mode (tk-name mode))  ;;;;  type is string 
    (if (equal? variation  'regular) 
	(set! bname type)
      ;; else 
	(set! bname (list variation type))
    )
    (foreach p ($ 'require bname) 
       ( if (not (tcl-info exists ($ 'poly p)))
           (eval  ($ 'build p))    ;;;  as (eval (init_hexahedron))
       )
    )
    (plot_poly path bname)  
    ((function path) ' eval :newlist dlist 'compile)
    (case mode 
	(( fill line)
	    ((function path) 'eval 
		:enable 'lighting 
		:disable 'stenciltest 
		:polygonmode 'frontandback mode )
	    (foreach dl ($ 'poly bname 'dlists)
		((function path) 'eval :call dl)
	    )
	 )
	((line/fill)
	    ((function path) 'eval  :enable 'stenciltest 
		:clear 'stencilbuffer 
		:stencilfunc 'always 0 1 
		:stencilop 'invert 'invert 'invert 
		:disable 'lighting 
		:polygonmode 'frontandback 'line)
	    (foreach dl ($ 'poly bname 'dlists) 
		((function path) 'eval :call dl 
		    :stencilfunc 'equal 0 1 
		    :stencilop 'keep 'keep 'keep 
		    :polygonmode 'frontandback 'fill 
                    :enable 'lighting
		    :call dl 
		    :stencilfunc 'always 0 1 
		    :stencilop 'invert 'invert 'invert 
		    :polygonmode 'frontandback 'line 
                    :disable 'lighting 
		    :call dl
                  )
	    )
	)
    )
    ((function path) ' eval :endlist)
)

(define ( redraw )   ;;;args  ???
    ;;global mode type variation dlist
    (polybuild .gl type variation dlist mode)
    (.gl 'redraw)
)

(define ( dumppoly  filename )
    ;;global poly bname
 (let* (( fd (open-output-file  filename )))
    (foreach f ($ poly bname 'face) 
	(write   (llength f) fd)
	(foreach v f 
	    (write  (lindex ( $ 'poly bname 'vtx) v) fd)
	)
    )
    (close-port  fd)
))
;;;;; global variables
(define dlist 0)
(define mode 0)
(define type  0)
(define variation  0)
(define bname  0)

(init_polybuild)
(pack (OGLwin '.gl :aspect 1 :stencil 8) :side 'left :fill 'both :expand 'yes)
(set! dlist (.gl 'newlist))

(pack (frame '.cmd) :side 'left :fill 'y)
(pack (frame '.cmd.main) :side 'top :fill 'x)

; quit button
(pack (button '.cmd.main.quit :text "Quit" :command "exit") :side 'left)

; mode menu
(set! mode 'fill)
(tk_optionMenu '.cmd.main.mode 'mode 'fill 'line 'line/fill)
;;;  trace variable mode w redraw     ******    to be  implement in scheme!!! 
(pack '.cmd.main.mode :side 'left)

; polygon type box
(pack (frame '.cmd.type :relief 'sunken :borderwidth 3) :fill 'x) 
(set! type 'hexahedron)
(foreach p '(tetrahedron hexahedron octahedron dodecahedron
               icosahedron) 
    (pack (radiobutton (tk-name '.cmd.type. p) :text p 
	      :var 'type :value p :command '(redraw))  :anchor 'w)
)

; polygon variation box
(pack (frame '.cmd.variation :relief 'sunken :borderwidth 3) :fill 'x )
(set! variation 'regular)
(foreach p '(regular stellated sunken half-truncated truncated) 
    (pack (radiobutton (tk-name '.cmd.variation. p) :text p :var 'variation
	     :value p :command '(redraw))  :anchor 'w)
)

; setup of initial OpenGL display
(polybuild .gl 'hexahedron 'regular dlist mode)
(.gl 'eval 
    :matrixmode 'projection 
    :loadidentity 
    :perspective 20 1 0.5 20
    :matrixmode 'modelview 
    :loadidentity 
    :lookat 0 0 10 0 0 0 0 1 0 
    :pushmatrix 
    :material 'front 'ambient 0.3 0.3 0.3 
    :material 'front 'diffuse 0.5 0.5 0.5 
    :material 'front 'specular 0.3 0.3 0.3 
    :material 'front 'shininess 40 
    :light 'light0 'position 1 1 1 0 
    :light 'light0 'ambient 1 1 1 
    :light 'light0 'diffuse 1 1 1 
    :light 'light0 'specular 1 1 1 
    :clearcolor 0 0 1 
    :enable 'lighting :enable 'light0 :enable 'light1 
    :enable 'depthtest 
)
(.gl 'main :clear 'colorbuffer 'depthbuffer 'stencilbuffer
    :call dlist) 

(rotation_bindings .gl)
(redraw)
;;(define (trace-mode)
;;  (tkwait 'variable 'mode)
;;  (redraw)
;;  (after 100 '(trace-mode))
;;)
;;(trace-mode) ;;; not complete!!!
(define (trace-mode x y z) (write x)(write z)(write mode)
     (after 100 '(redraw)))
;;;  after needs ,  this is a bug of TCL? see P.177 TCL/TK
;;;

(trace-variable 'mode 'w trace-mode)




