/****   Math  basic functions 
   f_math_ff.c 
***/
#define  F_MATH_FF_C

#include "lisp.h"
#include <stdio.h>
#include <math.h>

extern mpq_t mpqwork1, mpqwork2;
extern mpz_t mpzwork1, mpzwork2;
extern ptr genmul(),gendiv(),genadd(),gensub();
extern ptr fixdiv();
extern ptr DBL0;
extern int ODDP(ptr);   /* see biglog.c */

#define  getfloat(x)  (FLONUM(x))
extern double get_float();

/*  sin,cos,tan,acos,asin,atan,exp,sqrt */
ptr ATAN(),LOG(),EXP(),CIS();
ptr SINH(),COSH(), TANH(), ASINH(), ACOSH(), LOGRT();
static ptr mul_imag(y)
  Reg ptr y;
 {
    /*** return iy ***/
  if(REALP(y))
    return complex(ZERO,y);
  if(COMPLEXP(y))
    return complex(minus(CDR(y)),CAR(y));
  return error(E_notnum,y);
 }

static ptr mul_imag2(y)
  Reg ptr y;
 {
    /*** return -iy ***/
  if(REALP(y))
    return complex(ZERO,minus(y));
  if(COMPLEXP(y))
    return complex(CDR(y),minus(CAR(y)));
  return error(E_notnum,y);
 }
 
ptr ATAN(py,px)
 Reg ptr py,px;
{
 Reg double y,x;
 y=get_float(py);
 x=get_float(px);
      if(x==0.0)
       {
         if(y==0.0)  return error(E_divzero,py);
         if(y<0.0) return minus(ATOMVAL(a_pi2));
         return ATOMVAL(a_pi2);
       }
  y/=x;
  if(x>0.0) return MAKEDBL(atan(y));
  if(y<=0.0) return MAKEDBL(atan(y)+pi);
  return MAKEDBL(atan(y)-pi);
 }

ptr LOG(y)
    Reg ptr y;
 {
   Reg double x;
  if(zerop(y)) return error(E_illegal,y);
   if(REALP(y))
      {
          x=get_float(y);
         if(x<0.0)
           {
             return complex(MAKEDBL(log(-x)),ATOMVAL(a_pi));
           }
         return MAKEDBL(log(x));
       }
    if(COMPLEXP(y))
        {
           x=0.5*log(get_float(norm2(y)));
           y=ATAN(CDR(y),CAR(y));
           psh(y);
           y=complex(MAKEDBL(x),y);
           pull();
           return y;
        }
  return error(E_notnum,y);
 }

//  2018.3.14  this is OK!  (why???)
//ptr EXP(y)
ptr EXPGOOD(y)
    Reg ptr y;
 {
   Reg double x,b;
    Reg ptr z;
   if(REALP(y))
      {
   // fprintf(stderr,"EXPDebug real y\n");
          x=get_float(y);
        return MAKEDBL(exp(x));
       }
    if(COMPLEXP(y))
        {
           psh(y);
           x=exp(get_float(CAR(y)));
           z=MAKEDBL(x);
           psh(z);
           //b=get_float(CDR(y));
           y=CIS(CDR(y));
// y=MAKEDBL(x*cos(b));
           //psh(y);
           y=genmul(z,y);
//MAKEDBL(x*sin(b));
           //y=complex(y,z);
           stack-=2;
           return y;
        }
  return error(E_notnum,y);
 }

//  This cause Seg. fault  why???

//ptr EXP_NOGOOD(y)
//  2020.8.25  no problem
ptr EXP(y)
    Reg ptr y;
 {
   Reg double x,b;
    Reg ptr z;
   if(REALP(y))
      {
   // fprintf(stderr,"EXPDebug real y\n");
          x=get_float(y);
        return MAKEDBL(exp(x));
       }
    if(COMPLEXP(y))
        {
           psh(y);
           x=exp(get_float(CAR(y)));
           z=MAKEDBL(x);
           if(zerop(z)){
             pull();
             return CIS(CDR(y));
           }
           b=get_float(CDR(y));
           y=MAKEDBL(x*cos(b));
           psh(y);
           z=MAKEDBL(x*sin(b));
           stack-=2;
           if(zerop(z)){
               return y;
            }  
            return complex(y,z);
        }
  return error(E_notnum,y);
 }


ptr CIS(y)    /***  cos y+ i sin y  **/
    Reg ptr y;
 {
   Reg double x;
   if(REALP(y))
      {
          x=get_float(y);
          psh(MAKEDBL(cos(x)));
          y=MAKEDBL(sin(x));
        return complex(*pull(),y);
       }
  return error(E_notreal,y);
 }

ptr INTEXPT(a,y)
    Reg ptr a,y;    
 {
   /***   y ispos integer, a is any  ****/
    Reg ptr k,x;    
   if(y==ZERO) return ONE;
   if(y==ONE) return a;
   else if(y==TWO) return genmul(a,a);
   else
    {
     psh(k=fixdiv(y,TWO));   /**** 切り捨て整数。not rational!!! ***/
     psh(x=INTEXPT(a,k));
     *(stack-1)=genmul(x,x);
     if(ODDP(y)) *(stack-1)=genmul(*(stack-1),a);
     y=*pull();
     pull();
     return y;
   }
  }

ptr num_coerce(x,a)
  Reg ptr x,a;
 {
   if(INTEGERP(a)) return x;
   if(RATIOP(a)) return x;
   if(DOUBLEP(a)) return MAKEDBL(get_float(x));
   if(COMPLEXP(a)) 
   return complex(MAKEDBL(get_float(x)),DBL0);
   //return complex(MAKEDBL(get_float(x)),MAKEDBL(0.0));
  return error(E_notnum,a);
 }
   
     
ptr EXPT(a,y)
    Reg ptr a,y;
 {
   Reg double x;
   if(zerop(a)){
      /** printf("expt a== 0\n"); ***/
      if(zerop(y)) return num_coerce(ONE,a);
      return num_coerce(ZERO,a);
    }
     
   if(INTEGERP(y))
      {
        if(y==ZERO) 
               return  num_coerce(ONE,a);
            psh(y);
            psh(a);
             if(gsign(y)) 
              {
                *(stack-2)=y=minus(y);
                *(stack-1)=a=gendiv(ONE,a);
              }
             y=INTEXPT(a,y);
             sp_pull(2);
             return y;
     }
    psh(a=LOG(a));
    *(stack-1)=a=genmul(y,a);  
    a=EXP(a);             
    pull();
    return a;
 }

ptr s_exp(int narg,ptr env)
    
 {
  checknarg(narg-1);
    return EXP(*(stack-1));
 }

ptr s_expt(int narg,ptr env)
    
 {
  checknarg(narg-2);
    return EXPT(*(stack-2),*(stack-1));
 }

ptr s_cis(int narg,ptr env)
    
 {
  checknarg(narg-1);
    return CIS(*(stack-1));
 }
 
ptr s_atan(int narg,ptr env)
    
 {
   Reg ptr y,x;
   y=*(stack-narg);
   if(narg==2)
    {
      x=*(stack-1);
      if(x==ZERO||(get_float(x)==0.0))
       {
         if(y==ZERO||(get_float(y)==0.0))  return error(E_divzero,y);
         if(gsign(y)) return minus(ATOMVAL(a_pi2));
         return ATOMVAL(a_pi2);
       }
      return ATAN(y,x);
    }
   checknarg(narg-1);
   if(REALP(y))
       return MAKEDBL(atan(get_float(y)));
   if(COMPLEXP(y))
    {
      x=complexadd(y,ATOMVAL(a_imag));
      psh(x);
      y=complexsub(ATOMVAL(a_imag),y);
      psh(y);
      *(stack-1)=y=gendiv(x,y);
      *(stack-1)=y=LOG(y);
      *(stack-1)=y=mul_imag(y);
      y=gendiv(y,TWO);
      sp_pull(2);
      return y;
    }
  return error(E_notnum,y);
 }

ptr s_sqrt(int narg,ptr env)
    
 {
   Reg double x;
   Reg ptr y=*(stack-1);
   
  checknarg(narg-1);
   if(REALP(y))
     {
         if(gsign(y)) 
           return complex(ZERO,MAKEDBL(sqrt(-get_float(y))));
         return MAKEDBL(sqrt(get_float(y)));
     }
   if(COMPLEXP(y))
    {
       psh(y=LOG(y));
       *(stack-1)=y=gendiv(y,TWO);
       y=EXP(y);
       pull();
       return y;
    } 
  return error(E_notnum,y);
 }

ptr s_log(int narg,ptr env)
    
 {
   Reg ptr y,x;
    if(narg==1)
       return y=LOG(*(stack-1));
    checknarg(narg-2);
    
   x=*(stack-1);
      /*** x is base **/
       y=LOG(*(stack-2));
      psh(y);
      if(x==TWO) 
          x=ATOMVAL(a_log2);
      else if(x==TEN) 
          x=ATOMVAL(a_log10);
      else
          x=LOG(x);
      psh(x);
      y=gendiv(y,x);
      sp_pull(2);
    return y;
 }

ptr s_log10(int narg,ptr env)
    
 {
   Reg ptr y,x;
  checknarg(narg-1);
      y=LOG(*(stack-1));
      psh(y);
          x=ATOMVAL(a_log10);
      y=gendiv(y,x);
      pull();
   return y;
 }

ptr s_log2(int narg,ptr env)
    
 {
   Reg ptr y,x;
  checknarg(narg-1);
      y=LOG(*(stack-1));
      psh(y);
          x=ATOMVAL(a_log2);
      y=gendiv(y,x);
      pull();
   return y;
 }  


ptr s_sin(int narg,ptr env)
    
 {
   Reg ptr y=*(stack-1);
  checknarg(narg-1);
   if(REALP(y))
     {
    return MAKEDBL(sin(get_float(y)));
     }
   if(COMPLEXP(y))
    {
      psh(y=mul_imag(y));
      *(stack-1)=y=SINH(y);
      y=mul_imag2(y);
      pull();
     return y;
    }
  return error(E_notnum,y);
 }
ptr s_cos(int narg,ptr env)
    
 {
   Reg ptr y=*(stack-1);
  checknarg(narg-1);
   if(REALP(y))
     {
    return MAKEDBL(cos(get_float(y)));
     }
   if(COMPLEXP(y))
    {
      psh(y=mul_imag(y));
      y=COSH(y);
      pull();
     return y;
    }
  return error(E_notnum,y);
 }
ptr s_tan(int narg,ptr env)
    
 {
   Reg ptr y=*(stack-1);
  checknarg(narg-1);
   if(REALP(y))
     {
    return MAKEDBL(tan(get_float(y)));
     }
   if(COMPLEXP(y))
    {
      psh(y=mul_imag(y));
      *(stack-1)=y=TANH(y);
      y=mul_imag2(y);
      pull();
     return y;
    }
  return error(E_notnum,y);
 }
ptr s_asin(int narg,ptr env)
    
 {
   Reg ptr y=*(stack-1);
   Reg double x;
  checknarg(narg-1);
   if(REALP(y))
     {
        x=get_float(y);
       if(fabs(x)<=1.0)
         return MAKEDBL(asin(x));
       x+=sqrt(x*x-1.0);
       return complex(ATOMVAL(a_pi2),MAKEDBL(-log(x)));
     }
   if(COMPLEXP(y))
    {
      psh(y=mul_imag(y));
      *(stack-1)=y=ASINH(y);
      y=mul_imag2(y);
      pull();
     return y;
    }
  return error(E_notnum,y);
 }

ptr s_acos(int narg,ptr env)
    
 {
   Reg ptr y=*(stack-1);
   Reg double x;
  checknarg(narg-1);
   if(REALP(y))
     {
        x=get_float(y);
       if(fabs(x)<=1.0)
         return MAKEDBL(acos(x));
       x+=sqrt(x*x-1.0);
       return complex(ZERO,MAKEDBL(log(x)));
     }
   if(COMPLEXP(y))
    {
      psh(y=ACOSH(y));
      y=mul_imag2(y);
      pull();
     return y;
    }
  return error(E_notnum,y);
 }

double ss_sinh(x)
  Reg double x;
 {
  Reg double s,g,p;
  Reg int n;
   if(fabs(x)>0.3) return 0.5*(exp(x)-exp(-x));
   p=x*x;
   g=s=1.0;
   for(n=3;n<16;n+=2)
    {
      g*=p/n/(n-1);
      s+=g;
     }
   return s*x;
  }
double ss_cosh(x)
  Reg double x;
 {
    return 0.5*(exp(x)+exp(-x));
  }
ptr SINH(z)
   Reg ptr z;
 {
   Reg double x,y;
   if(REALP(z))
     {
        x=get_float(z);
         return MAKEDBL(ss_sinh(x));
     }
   if(COMPLEXP(z))
    {
       /***  cos y sinh x+i sin y cosh x  ***/
        x=get_float(CAR(z));
        y=get_float(CDR(z));
        psh(MAKEDBL(cos(y)*ss_sinh(x)));
        z=MAKEDBL(sin(y)*ss_cosh(x));
       return complex(*pull(),z);
     }
  return error(E_notnum,z);
 }
ptr COSH(z)
   Reg ptr z;
 {
   Reg double x,y;
   if(REALP(z))
     {
        x=get_float(z);
         return MAKEDBL(ss_cosh(x));
     }
   if(COMPLEXP(z))
    {
       /***  cos y sinh x+i sin y cosh x  ***/
        x=get_float(CAR(z));
        y=get_float(CDR(z));
        psh(MAKEDBL(cos(y)*ss_cosh(x)));
        z=MAKEDBL(sin(y)*ss_sinh(x));
       return complex(*pull(),z);
     }
  return error(E_notnum,z);
 }
       
ptr TANH(z)
   Reg ptr z;
 {
   psh(SINH(z));
   psh(z=COSH(z));
   z=gendiv(*(stack-2),z);
   sp_pull(2);
   return z;
 }
ptr ASINH(y)
   Reg ptr y;
 {
   Reg double x;
   Reg ptr *z;
   if(REALP(y))
     {
        x=get_float(y);
       x+=sqrt(x*x+1.0);
       return MAKEDBL(log(x));
     }
   if(COMPLEXP(y))
    {
      return LOGRT(y,ONE);
    }
  return error(E_notnum,y);
 }
ptr ACOSH(y)
   Reg ptr y;
 {
   Reg double x;
   Reg ptr *z;
   if(REALP(y))
     {
        x=get_float(y);
       if(fabs(x)>=1.0)
         {
           x+=sqrt(x*x-1.0);
           return MAKEDBL(log(x));
         }
      return LOGRT(y,-1);
     }
   if(COMPLEXP(y))
    {
      return LOGRT(y,-1);
    }
  return error(E_notnum,y);
 }

ptr LOGRT(y,c)
   Reg ptr y,c;
 {
   Reg ptr *z;
      z=l_sp;
      psh(genmul(y,y));
      *z=genadd(*z,c);
      *z=LOG(*z);
      *z=gendiv(*z,TWO);
      *z=EXP(*z);
      *z=genadd(y,*z);
      y=LOG(*z);
      pull();
     return y;
   }
ptr ATANH(y)
    Reg ptr y;
 {
   Reg double x;
   Reg ptr *z,w;
   if(NUMBERP(y))
     {
       if(REALP(y))
       {
        x=get_float(y);
       if(fabs(x)<=1.0)
         {
           x=(1.0+x)/(1.0-x);
           return MAKEDBL(0.5*log(x)); 
         }
        }
      psh(w=genadd(y,ONE));
      z=l_sp;
      y=gensub(ONE,y);
      psh(y);
      *z=gendiv(w,y);
      *z=LOG(*z);
      y=gendiv(*z,TWO);
      sp_pull(2);
      return y;
    }
  return error(E_notnum,y);
 }

ptr s_sinh(int narg,ptr env)
    
 {
    return SINH(*(stack-1));
 }
ptr s_cosh(int narg,ptr env)
    
 {
    return COSH(*(stack-1));
 }
ptr s_tanh(int narg,ptr env)
    
 {
    return TANH(*(stack-1));
 }
ptr s_asinh(int narg,ptr env)
    
 {
    return ASINH(*(stack-1));
 }
ptr s_acosh(int narg,ptr env)
    
 {
    return ACOSH(*(stack-1));
 }
ptr s_atanh(int narg,ptr env)
    
 {
    return ATANH(*(stack-1));
 }

 ptr s_isqrt(int narg,ptr env)
  {
    Reg ptr x,n,c,y,*sp;
    Reg double v;
    x=*(stack-1);
    if(SINTP(x)){
        v=get_float(x);
        v=sqrt(v);
        return MAKESINT((int)floor(v));
     }
    else if(BIGINTP(x)){
        mpz_sqrt(mpzwork1,BIGNUM(x));
        return mpz_check16(mpzwork1);
     }

    return error(E_notint,x);
}
 





