;;;  regexp.ss
;;;  regular expression  Library
;;;  2004.10.26
;;;  2004.10.30  automaton version  1.0
;;;  2004.11.8  VER1.2 beta

(in-package "regexp")

(define user-num 0)
(define user-group 0)
(define *nocase*  #f)
(define *group-res*  ())

;;;  :index -> 1  :repeat 4
(define (regexp str pat . flag)
   (set! flag (get-flag flag))
   (regexp-main (make-regexp  pat ) (string->list str) (string-length str) flag 0)
)

(define (get-flag flag)
  (let* ((f 0) (x 0))
     (set! *nocase*  #f)
     (while (pair? flag)
          (set! f (pop flag))
          (cond ( (eq? f :nocase)
                   (set! *nocase* #t))
                ((eq? f :index)  (set! x (+ x 1)))
                ((eq? f :repeat)  (set! x (+ x 4)))
                ((eq? f #t))
                (else (error 550 f))
          )
      )
   x
))

(define (regexp-group str pat . flag)
  (let* ((x 0)(size 0) (mf #t) (f 0)(start 0)(memo 0)(res ()))
     (set! size (string-length str))
     (set! str (string->list str))
     (set! memo str)
     (set! flag (get-flag flag))
     (set! pat (make-regexp pat))
     (if (= (logand flag 1) 1) (set! mf size) (set! mf #t))
     (if (= (logand flag 4) 0)
       (begin
          (set! x (regexp-group-sub str pat size mf 0))
          (set! *group-res* (list*  str flag x))
            x
       )
       (begin
         (while x
            (set! size (length str))
            (if (= (logand flag 1) 1) (set! mf size) (set! mf #t))
            (set! x (regexp-group-sub str pat size mf start))
            (when x
               (push x res)
               ;;;(format #t "gr:~s~%" x)
               (set! start (cdar x))
               (set! str (list-tail memo start))
            )
         ) 
         (if (pair? res)
            (begin 
              (set! res (reverse res))
              (set! *group-res* (list*  memo flag res))
              res
            )
            #f
          )
       )
    )
))


(define (regexp-group-sub  str pat size mf start)
  (let* ((x 0) (f 0))
     (set! x (regexp-top pat str size mf))
    ;;; (format #t "~s~%" x)
     (if x  
        (begin 
          (set! x (cons (+ start (car x)) (+ start (- size (length (cdr x))))))
          (if (and (> start 0)(number? mf))
                 (set! user-group (add-index user-group start)))
          (cons x user-group )
        )
        #f
     )
))

(define (add-index  data step)
   (map (lambda(x) (add-index-sub x step)) data)
)

(define (add-index-sub  data step)
 (let* ((x 0)(res ()))
   (while (pair? data)
       (set! x (pop data))
       (push (cons (+ step (car x))(+ step (cdr x))) res)
   )
  (reverse res)
))
 
(define (get-regexp-group n)
   (let* ((data  *group-res*) (str 0)(mf 0))
      (set! str (pop data))
      (set! mf  (pop data))
      (if (>= mf 4)
          (map   (lambda (x)  (get-group-sub (list-tail x n)  str mf)) data)
          (get-group-sub (list-tail data n)  str mf)
      )
))

(define (get-group-sub data str mf)
   (set! data (car data))
   (if (pair? data)
        (begin
       #|
          (if (or (= mf 1)(= mf 5))  ;;;  :index
               (set! str (map (lambda(x) (make-str-by str (car x) (cdr x))) data))
               (set! str data)
          )
      |#
         ;;; (if (= (length str) 1) (car str) str)
         data   ;;;str
       )
      #f
   )
)
     
(define (make-str-data from n)
     (set! from (car (list-tail from n)))
     (conn-str-space from)
)

(define (conn-str-space lis)
  (if (null?  lis)
        ""
     (if (= (length lis) 1)
        (car lis)
        (string-append (car s) " " (conn-str-space lis))
    )
  )
)


(define (regexp-subst str pat rstr . flag)
   (let* ((ind 0) (index 0) (opt 1) (res ())(from 0) (f 0) (subf #f))
     (set! opt (get-flag flag))
     (if (= opt 0) (set! opt 1))
     (set! index (apply regexp str pat :index flag))
     (if index
       (begin
          (if (not (string? rstr)) (set! subf rstr))
          (if (=  opt 1) (set! index (cons index ())))
          (while (pair? index)
             (set! ind (pop index))
             (push (substring str from (car ind)) res)
              (if subf
                  (set! rstr (funcall subf (substring str (car ind) (cdr ind))))
              )
             (push rstr res)
             (set! from (cdr ind))
          )
          (push (substring str from (string-length str)) res)
          (apply string-append (reverse res))
        )
        #f
    )
))
  
;;
(define (make-regexp pat)
  (let* ((s 0))
   (if (string? pat)
     (begin
      (set! user-num 0)
      (set! pat (string->list pat))
      (set! s (car (analy-pat-rec pat #f)))
      ;;;(format #t "analy:~s~%" s)      
      (set! s  (reg-new  s))
      (if (eq? (car s) 'user-group) (set! s (cons s ())))  ;;; case whole is (...)
      (cons user-num  s)
    )
   pat
  )
))

(define (r-esc x)
   (cond ((eq? x #\n) #\newline)
         ((eq? x #\t) #\tab)
         ((eq? x #\e) #\escape)
         ((eq? x #\f) #\page)
         ((eq? x #\r) #\return)
         ((eq? x #\s) 'r-space?)
         ((eq? x #\S) 'not-space?)
         ((eq? x #\d) 'r-digit?)
         ((eq? x #\D) 'not-digit?)
         ((eq? x #\w) 'r-word?)
         ((eq? x #\W) 'not-word?)
         ((eq? x #\k) 'r-kanji?)
         ((eq? x #\K) 'not-kanji?)
         ((eq? x #\/) #\\)
         (else x)
   )
)


(define (analy-pat-rec  pat gr)
  (let* ((x 0)(res ()) (sel #f)(cont #t) (k 0))
    (while (and (pair? pat) cont)
        (set! x (pop pat))
        (cond ((eq? x #\.)(push 'r-dot res))
              ( (eq? x #\*) (push (list :* (pop res)) res))
              ( (eq? x #\+) (push (list :+ (pop res)) res))
              ( (eq? x #\?) (push (list :? (pop res)) res))
              ((eq? x #\( )
                   (set! res (r-group-new  res))
                   (inc user-num)
                   (set! k user-num)
                   (set! x (analy-pat-rec pat k))
                   (push (caar x) res)
                   (set! pat (cdr x))
                  )
              ((eq? x #\\ )
                  (set! x (pop pat))
                  (push (r-esc x) res)
                  )
              ((and (eq? x #\)) gr )
                   (set! res (r-group-new  res))
                   (push gr res)
                   (set!  x (reverse res))
                   (set! x (cons gr x))
                   (set! x (cons 'user-group x))
                   (set! res (cons x ()))
                   (set! cont #f)
                  )
              ((eq? x #\[ )(set! x  (reg-bag pat))
                           (push (car x) res)
                           (set! pat (cdr x))  )
              ((eq? x #\^) (push 'r-top res))
              ((eq? x #\$) (push 'r-end res))
              ((eq? x #\|) 
                 (set! sel #t)
                 (set! res (r-group-new res))
                 (push 'r-select res))
              (else (push x res))
        )
    )
         (set! res (r-group-new  res))
         (cons (reverse res) pat)
))

(define (group-sepa? x)
  (or
    (eq? x 'r-select)
    (if  (pair? x)
       (let* ((top (car x)))
          (cond ((eq? top 'user-group) #t)
                ((keyword?  top)   ;;; :*,:+,:?
                   (and (pair? (cadr x)) (eq? (caadr x) 'user-group)) )
                (else #f)
          )
       )
      #f)
))
 

(define (r-group-new lis)
  (let* ((s ()) (x 0) (cont #t))
    (while (and (pair? lis) cont)
       (set! x (car lis))
       ;;;(if  (or (and (pair? x) (eq? (car x) 'user-group))  (eq? x 'r-select))
       (if  (group-sepa? x)
            (begin
               (set! cont #f)
               (when (pair? s)
                   (if (pair? (cdr s)) 
                       (push s lis)
                       (push (car s) lis)
                   )
                   (set! s ())
               )
            )
           (begin
               (set! lis (cdr lis))
               (push x s)
           )
       )
    )
   (if (pair? s)
         (if (pair? (cdr s))
             (push s lis)
             (push (car s) lis)
         )
   )
   lis
))

             
(define (reg-bag pat)
   (let* ((x 0)(y 0)(z 0)(res '(r-bag)) (cont #t))
     (set! x (pop pat))
     (when (eq? x #\^) 
          (push 'not res)
          (set! x (pop pat)) )
     (when (eq? x #\])
          (push x res)
          (set! x (pop pat))  )
      (if (eq? x #\])
            (cons (reverse res) pat)
        (begin
          (if (eq? x #\\)
              (push x pat)
              (push x res)
          )
          (while (and cont (pair? pat))
              (set! x (pop pat))
              (cond ((eq? x #\-)
                       (if (and (char? (car res)) (not (eq? (car pat) #\])))
                              (begin
                                (set! y (char->integer (pop res)))
                                (set! z (char->integer (pop pat)))
                                (push (list '<= y z) res)
                               )
                              (push x res)
                      )  )
                   ((eq? x #\])
                      (set! cont #f) )
                   ((eq? x #\\)
                      (push (r-esc (pop pat)) res)) 
                   (else  (push x res))
               )
           )
          (cons (reverse res ) pat)
        )
     )
))

(define (reg-new pat)  ;;; for pattern list
   (let* ((x 0) (res ())  (s 0) )
      (while (pair? pat)
         (set! x (pop pat))
         (if (pair? x)
           (push (reg-new-sub x) res)
           (if (eq? x 'r-select)
               (push (list :or (pop res) (reg-new-sub (pop pat)) ) res)
               (push x res)
           )
        )
    )
  (if (pair? res)
        (reverse res)
        pat
  )
))

               
(define (reg-new-sub  pat)  ;;;for start point of a pair or element
   (let* ((x 0) (memo pat)  (s 0) )
      (if (pair? pat)
        (begin
            (set! s (pop pat))
            (cond ((eq? s :*)  
                     (list :go (list (reg-new-sub  (car pat)) :back))   )
                  ((eq? s :+)    (list  (reg-new-sub (car pat)) :back)  )
                  ((eq? s :?)    (list :go (reg-new-sub  (car pat)))    )
                  ((eq? s 'user-group)
                      (cons s (reg-new pat))  )
                  ((eq? s 'r-bag)
                      memo)
                  (else 
                      (reg-new memo)  )
                 ;;; (else memo)
             )    
         )
        pat
     )
))

(define (regexp-main pat str size flag start)
   (let* ((x 0) (res () )(end 0)(end2 0))
     (set! x (regexp-top pat str size #f))
     (if x
        (begin
           (set! end (- size (length (cdr x))))
          (cond ((=  flag 0)   ;;; #t
                  (make-str-by str (car x) end  ) )
               ((= flag  1)   ;;;  :index
                  (cons (car x) end) )
               ( (= flag 4)    ;;; :repeat 
                     (set! res (make-str-by str (car x) end))  
                   (set! x (cdr x))
                   (set! x  (regexp-main pat x (length x) flag 0))
                   (if x
                       (cons res x)
                       (cons res ())
                   )
                 )
               ((=  flag 5 )  ;;;;  :index :repeat
                   (set! end2  (+ start end))
                   (set! res (cons (+ start (car x))  end2)) 
                   (set! x (cdr x))
                   (set! x  (regexp-main pat x (length x) flag  end2))
                   (if x
                       (cons res x)
                       (cons res ())
                   )
                 )
               (else (error 552 flag))
           )
        )
        #f
    )                   
))

(define (make-str-by str from end)
   (set! end (- end from))
   (while (> from 0)
      (set! str (cdr str)) (dec from))
   (if (= end (length str))
       (list->string str)
       (let* ((new ()))
          (while (> end 0)
             (push (pop str) new)
             (dec end)
          )
         (list->string (reverse new))
       )
   )
)

;;;  
(define (regexp-top pat str size mf)
   (let* ((memo #f)(ind 0) (end  #t) (v 0))
      (set! memo (r-auto str pat))
      (when memo
         (set! v (car memo))
         (if mf (set! user-group (make-user-data str v mf)))
         (set! ind (- size (length (car (vector-ref v 0)))))
         ;;(set! end (- size (length (car memo))))
         ;;(format #t "regexp:~s ~s~%" ind end)
       )
    (if memo
       (cons ind  (caddr memo))
       #f
    )
))

(define (make-user-data str v mf)
  (let* ((n (vector-length v))(i 1)(res ()) (x 0))
    (while (< i n)
     (if (number? mf)
          (set! x  (user-data-index  (vector-ref v i) mf))
          (set! x  (user-data (vector-ref v i)))
     )
      (push x res)
      (inc i)
    )
    (reverse res)
))

(define (user-data lis)
   (let* ((res ()) (s 0) (e 0))
     (while (pair? lis)
       (set! e (pop lis))
       (set! s (pop lis))
       (push (make-str-by  s 0 (- (length s) (length e))) res)
     )
   res
)) 

(define (user-data-index  lis size)
   (let* ((res ()) (s 0) (e 0))
     (while (pair? lis)
       (set! e (pop lis))
       (set! s (pop lis))
       (push (cons (- size (length s)) (- size  (length e))) res)
     )
   res
))
   
 ;; (format #t "~s~%" v)
  


(define r-space? 
  char-whitespace? )
 ;; (or (eq? x #\space) (eq? x #\tab)
 ;;                            (eq? x  #\newline)(eq? x #\page) (eq? x #\return))
;;)

(define (not-space? x) (not (r-space? x)))

(define (r-digit? x)
  (char-numeric? x)
)

(define (not-digit? x)
  (not (char-numeric? x))
)

(define r-word?     char-word?)
;;  (or (char-numeric? x) (char-alphabetic? x)(eq? x #\_))
;;;)

(define (not-word? x)
  (not (r-word? x))
)

(define (r-dot x)
  #t
)

(define r-kanji?   char-kanji?)
;;;  (>= (char->integer x) 256)
;;;)

(define (not-kanji? x)
  (< (char->integer x) 256)
)

(define (r-bag bag c)
  (let* ((x 0)(v 0)(not-f #f)(ans #f))
    (if (null? bag)
              (error 553 bag)    ;;; error
      (begin
        (if (eq? (car bag) 'not)
             (set!  not-f (pop bag)) )  
        (while (and (pair? bag) (not ans))
          (set! x (pop bag))
            (if (pair? x)
              (begin
                (set!  v (char->integer c))
                (set! ans (and (<= (cadr x) v) (<= v (caddr x))))
              ) 
             (if (char? x)
                 (set! ans 
                       (if *nocase* 
                             (char-ci=? x c)
                             (eq? x c)
                       )
                 )
                ;; else
                 (set! ans
                    (if(symbol? x ) (funcall x c)
                           (error 554 x)   ;;; sys error
                    )
                 )
             )

           )
       )
      (if not-f (not ans) ans)
     )
   )
))
;;;;   automaton version
;;;;   state= list of (pat user-vector parent-stack str)
;;;;   parent-stack is stack of (parent-node . str)
;;;;

(define (test str pat)
  (r-auto (string->list str) pat)
)
        
(define (r-auto str pat)
  (let* ((ugroup (pop pat)) (s 0)(ans #f)(v 0)(org str) (ct 2))
       (set! v (make-vector (+ ugroup 1) ()))

    (while (and (not ans) (> ct 0))
       (set! s (list pat  v () str))
       (vector-fill! v ())
       (push-vec v 0 org)
       (push-vec v 0 str)
       (set! s (do-auto-main (list s) (length str) ))
       (if  s 
            (set! ans  s)
           (begin
             (if (pair? str)
                  (set! str (cdr str))
                  (dec ct)
             )
           )
       )
    )
  ans
))

(define (do-auto-main s-list size )
  (let* ((ans #f) (c 0) (s 0)(x 0) (res ()) (len (+ size 1)))
   (while  (pair? s-list)
      (set! res ())
      (while (pair? s-list)
         (set! s (do-auto  (pop s-list) )) 
         (while (and (pair? s) (> len 0))
            (set! x (pop s))
           ;;; (push x res)
            (if  (eq? (car x) ())
                (begin
                   (set! x (cdr x))
                   (when (< (length (caddr x)) len)
                      (set! ans x)
                      (set! len (length (caddr x)))
                   )
                  ;;; more
                 )
               (push x res)
           )
         )
      )
      (set! s-list res)
   )
  ans
))


(define (do-auto   s )
  (let* ((pat (car s)) (str 0)  (v 0) (par 0)(op 0)(p-op 0))
     ;;;(format #t"do-auto pat ~s v ~s par ~s str ~s~%" pat (cadr s)(caddr s)(cadddr s))
           (set! v (cadr s))
           (set! par (caddr s))
           (set! str (cadddr s))
     (if (null? pat) 
          (list s)
        (begin
           (set! op (car pat) )
           (cond ((symbol? op)
                      (auto-symbol op (cdr pat) v par str)  )
                ((char? op)
                     (auto-char op (cdr pat)  v par str)  )
                ((number? op)
                     (auto-user op (cdr pat)  v par str)  )
                ((pair? op)
                      (auto-pair op pat   v par str)  )
                ((eq? op :back)
                    ;;; (format #t ":back pat ~s par ~s~%" pat par)
                    ;;; (read)
                     (back-node-rec (cdr pat) v par str))
                (else (error 556 op))
           )
       )
   ) 
))

(define (back-node pat  v par  str)
 (let* ((node 0) )
  (if (and (null? pat) (pair? par))
        (begin
             (set! node (pop par))
             (set! pat (cdar node))
             (back-node pat v par str) ;;; pat may be ()
        )
      (list( list pat v par str))
  )
))

(define (back-node-rec  pat v par str)
 (let* ((node 0)  (s 0))
  (if (null? pat)
      (if (pair? par)
           (begin
             (set! node (pop par))
             (set! pat (car node))
             (set! s  (back-node (cdr pat) v par str))
              (if  (eq? str (cdr node))
                    s
                   (list (car s) (car (back-node pat (vector-copy v) par str))) 
              )
           )
       )
      (back-node pat v par str)
  )
))

(define (auto-symbol op pat  v par str)
  (cond ((eq? op 'r-top)
            (if (equal? str (cadr (vector-ref v 0)))
                 (back-node pat  v par str)
                #f
            )  )
        ((eq? op 'r-end)
           (if (null? str)
                 (back-node pat  v par str)
                #f
           )  )
        (else
           (if (and (pair? str) (funcall op (pop str)))
                 (back-node pat  v par str)
               #f
           )
        )
  )
) 
        
(define (auto-user op pat  v par str)
    (push-vec v op str)  ;;;  v 
      ;;; (format #t "user-end: par ~s pat ~s v ~s~%" par pat v)
    (back-node pat  v  par str)
)

(define (auto-char c pat  v par str)
  (and (pair? str)
         (if *nocase* (char-ci=? c (pop str))
                      (eq? c (pop str))
         )
             (back-node pat  v par str)
  )
)

(define (auto-pair op pat  v par str)
  (let* ((top (car op))(user 0) )
    (cond ((eq? top 'user-group)
             (push (cons pat str) par)
             (set! user (cadr op))
             (set! pat (cddr op))
             (push-vec v user str)  ;;;  v
             ;;;(format #t "user-start: par ~s pat ~s v ~s~%" par pat v)
             ;;; (debug)
               (back-node pat v par str)
              )
         ((keyword? top)
             (push (cons pat str) par)
              (auto-key top (cdr op) pat  v par str) 
              )
         ((eq? top 'r-bag)
              (if (and (pair? str) (r-bag (cdr op) (car str)))
                   (back-node (cdr pat) v  par (cdr str) )
                  #f
              )
           )
         (else
             (push (cons pat str) par)
            (set! pat  op ) 
            (back-node pat v par str)
         )
     )
))

(define (auto-key top op pat  v par str)
  (let* ((memo #f) (s 0) (n (length str)) (org 0)(i  0) (new 0))
    (cond ((eq? top :or)
             (set! s (list (car op)))
             (set! s (list s  v par  str ))
             (set! new (list (cadr op)))
             (set! new (list new  (vector-copy v)  par  str ))
             (list s new)
            )
         ((eq? top :go)   ;;;  == (:or () (car op))
             (set! s (back-node (cdr pat) v (cdr par)  str))
             (set! new  (list (car op)))
             (set! new (list new  (vector-copy v)  par  str ))
             (list (car s) new)
           )
         (else
             (error 557 top)  )
    )
))
        
(define (push-vec vec n val)
  (let* ((x (vector-ref vec n)))
     (vector-set! vec n (cons val x))
))

(set! *auto-regexp* make-regexp)
(define ss:regexp regexp:regexp)    ;;;  ss:regexp =  tk:regexp overwrite!
(export '(regexp-group regexp-subst get-regexp-group make-regexp))
(use-package "regexp" "ss")                  
        

