;;;;  2005.1.19
;;;;   html analysis
;;;;   like an automaton
;;;;   accept char
;;;;   eof is  ()

;;;   world is the  h-state
;;;;   h-state  is a (state . res-stack)
;;;    state is  stack of 
;;;;   auto-<  auto-tag  auto-tag/  auto-string auto-word auto-text
;;;;   auto-tag-p 
;;;    auto-tag! auto-tag!- auto-tag!--  
;;;;  if error continue as auto-text

;;;    (char .  h-state) means ungetchar
;;  2005.2.23
;;  New STATE version 

(in-package "http")

(define *pack* (find-package "http"))
(define *state* 0)  ;;; for debug

(defstruct STATE (:state (list auto-text)) (:result ())
        (:ungetc #f)
        (:out (open-output-str "html"))
        (:subout (open-output-str "html"))
)

(define (world-begin)    ;;;first world
   (set! *state*  (make-STATE))
   *state*
)

(define (out-flush state)
   (let* ((out (STATE-out state)) (s (close-port out)) )
     (reopen-str out :output)
     (set-STATE state :result (cons s (STATE-result state)))
   state
))

(define (subout-close state)
   (let* ((out (STATE-subout state)) (s (close-port out)) )
     (reopen-str out :output)
    s
))

(define (out-close state)
   (let* ((out (STATE-out state)) (s (close-port out)) )
    s
))

(define (out-open state)
   (let* ((out (STATE-out state)) )
     (reopen-str out :output)
))

(define (out-append  state)
   (let* ((out (STATE-out state)) )
     (reopen-str out :append)
))

(define (push-result x state)
   (let* ((s (STATE-result  state)))
     (set-STATE state   :result (cons x s))
   state
))

(define (push-state x state)
   (let* ((s (STATE-state  state)))
     (set-STATE state   :state (cons x s))
    state
))

(define (pull-state  state)
   (let* ((s (STATE-state  state)))
     (set-STATE state   :state (cdr  s))
    state
))


(define (go-state x state)
   (let* ((s (STATE-state  state)))
     (set-STATE state   :state (cons x (cdr s)))
   state
))

(define (push-char ch state)
    (display ch (STATE-out state))
  state
)

(define (push-char-sub ch state)
    (display ch (STATE-subout state))
  state
)

      
(define (html-auto ch world)
 (let* ((state 0) (res 0 ) (s #f))
   (if (null? ch)
     (begin
        (set! state (STATE-state world))
        (set! res (STATE-result  world))
        (set! s (STATE-out  world))
     
        (when (not (closed-port? s))
           (set! s (close-port s)) 
           (if (not (equal? s "")) 
               (push s res)
           )
        )
       (reverse res)  ;;;;   end auto html
     )
    (begin
       (when (not (eq? ch #\tab))   ;;;; always ignore tab
         (set! world (html-auto-exec ch world))
         (set! s (STATE-ungetc world))
        )
       (if s  ;;;;UNGETCHAR   
           (begin
             (set-STATE world :ungetc #f)
             (html-auto-exec s world)
           )
           world
       )
    )
  )
))

(define (html-auto-exec  ch world)
  (let* ((x #f))
     (set! x (html-auto-state ch world))
     (if x
             x
        (begin ;;; error
            (push-state auto-text world)
            (push-result  :error world)
            (out-flush world)
            world
        )
     )
))

(define (auto-skip-space ch world)
   (when (not (eq? ch #\space))
       (pull-state world)
       (set-STATE world :ungetc ch)
   )
 world
)

(define (html-auto-state ch hstate)
  (let* ((state (STATE-state hstate))
         (s (car state)) )
       (funcall s ch hstate )
))

(define (auto-< ch state )
    (cond ( (eq? ch #\/)
                 (go-state  auto-tag/  state)
                   )
          ((char-whitespace?  ch)  state )
          ((char-word? ch)
                (go-state  auto-tag  state)
                (push-result  :tag state)
                (auto-word-start  ch state ) 
              )
          ((eq? ch #\!)
                 (go-state  'auto-tag!  state)
              )
          (else  #f)    ;;; (error 666 state)   
     )
)

;;;; text
(define (auto-text ch state )
   (cond ( (or (eq? ch #\<) (eq? ch #\")(eq? ch #\newline))
             (auto-text-close ch state )  )
         ( (eq? ch #\&)
             (push-state  auto-entity state)    
           )
         (else
             (push-char  ch state)
          )
   )
)

(define (auto-text-close ch state )
  (let* ((s 0)(out (STATE-out state)))
   (out-flush state)
   (cond ( (eq? ch #\<)
             (push-state  auto-<  state))
         ( (eq? ch #\")
            (push-state   auto-string state)
           )
         ((eq? ch #\newline)
           (push-result  #\newline state)
            )
        (else  )    ;;; never ocuur
   )
    state 
))

(define (auto-entity ch state )
  (let* ((s 0)(memo 0)(c #f)(ungetc #f))
   (if (or (eq? ch #\;) (eq? ch #\space)(eq? ch #\<)(eq? ch #\"))
      (begin
         (set! s (subout-close state))
          (if (eq? ch #\;)
              (set! c (entity->char s) )
              (set! ungetc ch)   ;;;;   UNGETC
          )
          (if c
             (push-char  c state)
             (begin
               (push-char #\& state)
               (push-char s  state)
             )
          )
          (pull-state state)
          (if ungetc (set-STATE state :ungetc ungetc)) 
     )
    (push-char-sub   ch state)
  )
  state
))

(define  (entity->char ent) 
   (cond ((string-ci=? ent "amp") #\&)
                 ((string-ci=? ent "copy") #\@)
                 ((string-ci=? ent "lt") #\<)
                 ((string-ci=? ent "gt") #\>)
                 ((string-ci=? ent "nbsp") #\space)
                 ((string-ci=? ent "quot") #\" )
                 (else #f)
           )
)

;;;;;  tag  this apeears after read a word
(define (to-lower x)
   (cond ( (string? x) x)
         ( (pair? x)
             (if (eq? (car x) :word) 
                   (string-downcase (cdr x))
                   :error
             )
           )
         (else  :error)
   )
)

(define (ss-intern x)
   (let* ((s (find-symbol x)))
     (if s
         s   ;;; already in ss
         (intern x *pack*)
     )
))
        
(define (name-lower x)
  (set! x (to-lower x))
  (if (string? x)
       (ss-intern x )
       x
  )
)

(define (tag-name  pre str)
  (ss-intern (string-append pre str) )
)

(define (make-tag-prop res)
  (let* ((x 0)(xv 0)(prop ()) (cont #t)(name 0))
    (while (and cont (pair? res))
      (set! x (pop res))
    ;;;  (if  (null? x) (format #t "NULL in tag stack ~s~%" (cadr res)))
      (when (not (null? x))
        (set! xv (to-lower x))
        (if (or (eq? x :tag) (eq? xv :error))
           (begin
            ;;; (format #t "~s ~s~%" x xv)
             (set! cont #f)
           )
          (if (eq? (car res) #\=)
              (begin
                 (set! res (cdr res))
                 (set! name (name-lower (pop res)))
                 (push (cons name xv) prop)
              )
             (if (eq? (car res) :tag)
                 (push  (tag-name "t-" xv) prop)
                 (push (cons (ss-intern xv ) #f) prop)
             )
          )
       )
     )
    )
  (cons prop res)
   ;;;;(cons (cons :tag prop) res)
)) 
                  
(define   (auto-tag ch state )  
  (let* ((res (STATE-result state)))
  (cond ( (eq? ch #\>)
            ;;;(pull-state  state )
            (go-state auto-skip-space  state )
           ;;; (push :tag-end res)  
             (set-STATE state :result  (make-tag-prop res))
             )
        ( (eq? ch #\")
              (push-state   auto-string state)
             ;;; (push () res)  
            )
        ( (char-whitespace?  ch )   )
        ( (eq? ch #\=)
           ;;;  (out-flush state)
             (push-result  ch state)
             (go-state auto-tag-value-start state)
           )
        ( else
           (push-state  auto-word state)
           (push-char ch  state)
        )
  )
   state 
))

(define   (auto-tag-value-start  ch state )
  (let* ((res (STATE-result state)))
    (cond ( (eq? ch #\>)  ;;; error  but appears in google answer as <font color=>
            ;;;(pull-state  state )
           ;;; (push :tag-end res)
             ;;;(push-result   "" state)
             (push "" res)
             (set-STATE state :result (make-tag-prop res))
               (go-state auto-skip-space state )
             )
        ( (eq? ch #\")
              (go-state auto-tag-value state)
              (push-state  auto-string state)
              )
        ( (char-whitespace?  ch )   )
        ( (eq? ch #\=)    ;;;; error
             (push-result  :error state)
             (pull-state  state)
             )
        ( else
           (go-state auto-tag-value state)
           (push-state auto-value state)
           (push-char ch  state)
        )
  )
   state
))


(define   (auto-tag-value  ch state )
  (let* ((ex #f)(res (STATE-result state)))
  (cond ( (eq? ch #\>)
           ;;; (pull-state state)
             (set-STATE state :result (make-tag-prop res))
             (go-state auto-skip-space state )
             )
        ( (eq? ch #\")   ;;;; error
              (go-state auto-tag state)
              (push-state   auto-string state)
                )
        ( (char-whitespace?  ch ) 
              (go-state auto-tag state)
           )
        ( (eq? ch #\=)    ;;;; error
             (push-result :error state)
             (pull-state  state)  
             )
        ( else
            (go-state auto-tag state)
            (set! ex (auto-tag ch state ))
        )
  )
  (if ex ex  state)
))


(define (make-tag/ res)
  (let* ((x 0)(xv 0)(prop ()) (cont #t)(name 0))
    (while (and cont (pair? res))
      (set! x (pop res))
      (when (not (null? x))
        (set! xv (to-lower x))
        (if (or (eq? x :tag/) (eq? xv :error))
           (begin
             ;;;(format #t "~s ~s~%" x xv)
             (set! cont #f)
           )
          (if (eq? (car res) :tag/)
               (push (tag-name "t/" xv)  prop)
              (push  xv prop)   ;;;;  error!!!
          )
       )
     )
    )
   (cons (cons :tag/ prop) res)
   ;;;(cons  prop res)
))


(define   (auto-tag/ ch state )
 (let* ((s #t))
   (cond 
         ( (char-whitespace?   ch )   )
         ( (char-word? ch)
            (go-state auto-tag/word state)
           (push-state auto-word state)
            (push-result  :tag/ state)
           (push-char ch state)
          )
         (else  (set! s #f))  ;;;; error

   )
  (if s
      state 
     #f
  )
))


(define   (auto-tag/word ch state )
 (let* ((s #t) (res (STATE-result state)))
   (cond ( (eq? ch #\>)
            ;;;(pull-state  state )
            ;;;;(push :tag-end  res)
             (set-STATE state :result (make-tag/ res))
              (go-state auto-skip-space state)
             )
         ( (char-whitespace?   ch )   )
         (else  (set! s #f))  ;;;; error

   )
  (if s
     state 
     #f
  )
))


(define (auto-string ch state )
  (let* ((s 0))
   (cond ( (eq? ch #\")
            (pull-state  state )
            (out-flush state)
              )
         (else   
           (push-char ch  state)
         )
   )
  state 
))

(define (auto-word-start  ch state )
      (push-char ch  state)
     (push-state  auto-word state)
      state 
)

(define (auto-word ch state )
  (let* ((s 0))
   (cond ( (char-word?  ch)
            (push-char ch  state)
            state
            )
         (else
            (pull-state state )
            (set! s (out-close state))
            (set! s (cons :word s))
            (push-result s state)
            (out-open state)
            (funcall (car (STATE-state state)) ch state )
         )
   )
))

(define (auto-value ch state )
  (let* ((s 0))
   (if (or (char-whitespace? ch) (eq? ch #\>))
       (begin
            (pull-state  state )
            (set! s (out-close state))
            (set! s (cons :word s))
            (push-result s state)
            (out-open state)
            (funcall (car (STATE-state state)) ch state )
       )
      (begin
           (push-char ch state)
           state
      )
  )
))

;;;;    comment-tag
(define (auto-tag! ch state )
  (let* ((s 0))
    (if (eq? ch #\-)
         (begin
           ;;;;(push-result :comment state)
           (go-state auto-tag!- state)
         )
        (begin     ;;; as <!doctype 
          ;;;(format #t "<! ???  ~s ~s ~s ~%" ch state res)
          (go-state 'auto-tag state)
          ;;;(set! res (cdr res))
          (push-result  :tag state)
          (push-char #\! state)
          (push-char ch state)
          (push-state  auto-word state)
        )
    )
  state
))

(define (auto-tag!- ch state )
    (if (eq? ch #\-)
        (go-state  auto-tag!-- state)
        #f
    )
)

(define (auto-tag!-- ch state )
  (let* ((s 0)(end 0))
    (if (eq? ch #\>)
         (begin
           (set! s (out-close state))
           (if (and (> (string-length s) 1)(equal? (substring s :end -2 :end) "--"))
              (begin
                 (set! s (substring s 0 :end -2))
                 (out-open state)
                 (push-result  (list :comment s) state)
                ;;; (push :comment-end res)
                 ;;;(pull-state state)
                 (go-state  auto-skip-space state)
             )
             (begin
               (out-append state)
               (push-char ch  state)
             )
           )
         )
       (push-char ch  state)
    )
   state 
))

                
;;;;  main program

(define  (html-analy  str-lis)
  (let* ((w 0)(x 0)  (s 0) (str 0))
    (set! w (world-begin))
    (while (pair? str-lis)
       (set! str (pop str-lis))
       (set! s (string->list str))
      (while (pair? s)
        (set! x (pop s))
        (set! w (html-auto x w))
       ;; (format #t "ch  ~s~%" x)
       ;;;  (format #t "world state ~s  res ~s ~%" (car w)(cdr w))
      )
     (set! w (html-auto #\newline  w))
    )   
     (set! w (html-auto () w))
    ;;; (format #t "world res ~s ~%" w)
   w
))
