;;;  regexp2.0
;;;   2005.8.19
;;;
;;;  reduce rules
;;;   :e  == epsilon transition
;;;  

(in-package "regexp")

(load "R-string")
;;;(load "R-reduce")
(load "R-nfa")
(load "R-gmode")
(load "R-subr")
;;;(load "R-minct")
(load "R-shift")
(load "R-nfaexec")
(load "R-read")  ;;; after nfaexec is safe Why????
(load "R-reduce")
(load "R-reduce-min")
(load "R-minct")
(load "R-debug")

(define *regexp-repeat*  #f)
(define *regexp-pos*  #f)
(define regexp-char=?  eq?)
(define *group-res*  ())
;;;; New
(define *regexp-shift*  #t)
(define *regexp-preshift*  #f)
(define *regexp-endshift*  #f)
(define *regexp-strshift*  #f)
(define *regexp-minct*  0)

;;;;  mode
(define *regexp-mode*  #f)
;;;  :index -> 1    group=8

(define (get-flag flag)
  (let* ((f 0) (x 0))
     (set! *regexp-repeat*  #f)
     (set!  *regexp-mode*  #f)
     (set!  *regexp-pos*  #f)
     (set! regexp-char=?  eq?)
     (R-reset-debug)

     (while (pair? flag)
          (set! f (pop flag))
          (cond ( (eq? f :nocase)
                   (set! regexp-char=?  char-ci=?)
                   )
                ((eq? f :index)  (set! x (logior x 1)))
                ((eq? f :repeat)  (set! *regexp-repeat* #t))
                ((eq? f :single)  (set! *regexp-mode* f))
                ((eq? f :multiple)  (set! *regexp-mode* f))
                ((eq? f :cpos)  (set! *regexp-pos* #t))
                ((eq? f :debug)  (R-set-debug))
                ((eq? f #t))
                (else (error 550 f))
          )
      )
   x
))

(define (init-regexp-start b)  ;;; optimize by top exp
    (let* ((p (car b)) (mi (car p)) (s (cdr p)))
       ;;;;(format #t "start ~s G-pos ~s mi ~s~%" b *G-pos* mi)
       (set! *regexp-shift*  #t)
       (set!  *regexp-minct*  mi)
       (set!  *regexp-preshift*  #f)
       (when s
          (cond ((eq? s 'abs-top) 
                     (set! *regexp-shift* #f))
                ((eq? s 'top)
                     (set!  *regexp-preshift*  R-shift-top) )
                (  (eq? s 'g-pos) 
                     (set!  *regexp-preshift*  R-shift-g-pos) )
                ( (eq? s 'G-pos)
                     (set! *regexp-shift* #f)
                     (set!  *regexp-preshift*  R-shift-G-pos) )
                ((eq? s 'abs-end)
                     (set! *regexp-shift* #f)
                     (set!  *regexp-preshift*  R-shift-abs-end) 
                     )
                ((eq? s 'end)
                     (set!  *regexp-preshift*  R-shift-end) 
                )
                (else
                  (format #t "starterror  ~s ~s~% " b s)
                )
           )
      )
   (init-regexp-end (cadr b))  ;;; optimize by end exp
   (init-regexp-str (caddr b))  ;;; optimize by end exp
))

(define (init-regexp-end b)  ;;; optimize by end exp
    (let* ((ma (car b)) (s (cdr b)) )
      ;;; (format #t "end ~s ma ~s~%" b ma)
       (set!  *regexp-endshift*  #f)
       (when (and ma s)
          (cond  ((eq? s 'abs-end)
                     (set!  *regexp-endshift*  
                          (lambda(i)(L-shift-abs-end i ma))
                     )
                    )
                ((eq? s 'end)
                    (set!  *regexp-endshift*
                          (lambda(i)(L-shift-end i ma))
                     )
                    )
                (else
                  (format #t "end error  ~s ~s~% " b s)
                )
           )
      )
))

(define (init-regexp-str b)  ;;; optimize by pure-str exp
   (if b
      (set!  *regexp-strshift*  
             (lambda(i)(R-shift-str i b)) 
      )
      (set! *regexp-strshift*  #f)
  )
)
     
(define (init-regexp str exp)
  (let* ((v 0)(b 0))
     (if (and (not (eq? str *regexp-string*)) (string? *regexp-string*))
         (set! *G-pos* 0)
    )
    (if (string? exp) (set! exp (regexp->nfa exp)))
    (set! b (car exp))
    (set! *group-num* (pop b))
    (set! *named-group* (pop b))
    (init-regexp-start b)
    (set! v  (cdr exp))
   (if (not (vector? v))
        (set! v (nfa v))
    )
    ;;;(set! b (car v))
    (set! *regexp-nfa*  v)
    (set! *regexp-string* str)
    (set! *regexp-length* (string-length str))
    (set! *group-res* #f)
))

(define (regexp-pos . para)
    (if (pair? para)
        (begin
          (set! *G-pos* (car para))
          (set! *regexp-string* #t)
        )
        *G-pos*
    )
)

;;;;  state is #(Rstate index group-stack start-pos)    
(define *G-string* #f) 
(define *G-pos* 0)
(define *regexp-string* "")
(define *regexp-length* 0)
(define *regexp-nfa* 0)
(define *regexp-state* (make-vector 4 #f))   ;;; always first state&&last
;;;;;

(define (regexp str pat . flag)
   (set! flag (get-flag flag))
   (init-regexp str pat)
   (regexp-exec  flag )
)

(define (regexp-exec  flag )
   (let* ((ans #f)(grans #f)(res ()) (group ())
           (ind (= (logand flag 1)1)) (rep *regexp-repeat*))
      (set! ans (regexp-main 0))
      (if rep
        (begin
          (while  ans
            (push (copy-state ans) res)
            ;;;(push (R-group-get ans ind) group)
            (push (R-group-get ans ) group)
            (set! ans (regexp-main (Rindex ans)))
          ) 
            (set! grans  (reverse group))
          (set! ans 
             (if ind 
                 (map  state->index (reverse res))
                 (map  state->string (reverse res))
              )
           )
       )
       (if ans
         (begin
          ;;; (set! grans (R-group-get  ans ind))
           (set! grans (R-group-get  ans ))
           (set! ans 
             (if ind
                (state->index ans)
                (state->string ans)
              )
           )
         )
         #f
       )
     )
  (if grans (set! *group-res*  grans))
  (if (= (logand flag 8) 8)    ;;; return group
         grans
         ans
  )
))

;;;;  start from i
(define (regexp-main i )
  (let* ((ans #f)(owari (- *regexp-length* *regexp-minct*)))
       ;;; (format #t "main start i= ~s~%" i)
     (when *regexp-preshift*
        (set! i (funcall *regexp-preshift* i))
        ;;;(format #t "main shift i ~s~%" i)
      )
      (when (and i *regexp-endshift*)
        (set! i (funcall *regexp-endshift* i))
        ;;;;(format #t "end  shift i ~s~%" i)
      )
      (when  (and i *regexp-strshift*  )
          (set! i (funcall *regexp-strshift* i))
        ;;;(format #t "str shift i ~s~%" i)
      )
     (when (and i (<= i owari))
       (if *regexp-shift*
         ;;;(while (and (not ans) (<= i *regexp-length*))  ;; =   is for #"$"  
         (while (and (not ans) (<= i owari))  ;; =   is for #"$"  
            (set! ans  (regexp-nfa  (init-state *regexp-state* i)))
             (inc i)    
         )
          (set! ans  (regexp-nfa  (init-state *regexp-state* i)))
       )
     )
  (if ans
       (begin
          (set! *G-pos* (Rindex ans))
         ;;; (set! *G-string* *regexp-string*)
          ans
       )
      (begin
         (if (not *regexp-pos*)
                 (set! *G-pos* 0)
         )
          #f
      )
  )
))
;;;  access functions
(define (get-trans Rstate)
    (vector-ref *regexp-nfa* Rstate)
)

(define (name->group name)
   (let* ((x (assq  name *named-group*)))
         (if x
             (cdr x)
             #f
         )
))


(define (set-Rstate state Rstate)
   (vector-set! state 0 Rstate)
)


(define (Rstate state)
   (vector-ref state 0)
)

(define (Rindex state)
   (vector-ref state 1)
)


(define (set-Rindex state ind)
   (vector-set! state 1 ind)
)


(define (push-group state gr v)
   (vector-set! state 2  (cons (cons gr v) (vector-ref state 2)))
)

(define (get-group state)
     (vector-ref state 2)
)

(define (set-group state gr)
     (vector-set! state 2 gr)
)


(define (get-start-pos state)
     (vector-ref state 3)
)

(define (state->index state)
         (cons (get-start-pos state) (Rindex state))
)

(define (state->string state)
    (R-substring  (get-start-pos state) (Rindex state))
)

;;;;;;

(define (init-state v i)
    (vector-set! v 0 1)   ;;;Rstate     Now not used!
    (vector-set! v 1 i)    ;;; index
    (vector-set! v 2 ())  ;;;;;group-stack
    (vector-set! v 3 i)    ;;;; Not change preserve start-pos
    v
)


(define (copy-state state )
   (vector-copy state)
)

(define (R-call op state) 
   (let* (( f (pop op)))
      ;;;(format #t "R-call f ~s op ~s ~%" f op)
       (set! op (map
                  ;;;(lambda(x) (if (or (number? x)(string? x))
                  (lambda(x) (if (or (number? x)(symbol? x))
                        (str-group-ref x state) x)) op))
        (apply (function f) op)
 ))


(define (regexp-nfa-atom trans state)
  (let* ((memo trans)(A (pop trans))(i 0)(ind 0)(op 0))
    (cond ((eq? A  'string)
              (regexp-string trans *regexp-string* *regexp-length* state))
          ((eq? A  'pure-string)
              (regexp-pure-string trans *regexp-string* *regexp-length* state))
          ((eq? A 'g-top)
                (set! i (car trans))
                (set! ind  (Rindex state))
                    (if (> i 0)  (push-group state memo (Rindex state)))
                state)
          ((eq? A 'g-end) 
                (set! i (car trans))
                    (if (> i 0)  (push-group state memo (Rindex state)) )
                 state)
          ((eq? A 'back-init)
                 (set! op (car trans))
                  (set-car! op -1)
                  state   ;;; toriaezu
              )
          ((eq? A 'back)    ;;;; *,+ 
                (set! i (car trans))
                (set! ind  (Rindex state))
                ;;;(format #t "back ~s ~s~%" (back-index state i) ind)
                (if (= (car i) ind)
                    #f   ;;; infinity loop 
                   (begin
                    (set-car! i ind)
                    state
                   )
               )
            )
         ((eq? A 'r-pos)
               (r-pos (car trans) state) 
           )
         ((eq? A 'G-pos)
             (if (= (Rindex state) *G-pos*) state #f)
            )
         ((eq? A 'g-pos)
             (if (>= (Rindex state) *G-pos*) state #f)
            )
         ((eq? A 'group-ref)
              (group-ref (car trans) state) 
           )
         ((eq? A 'nfa-ref)
                ;;;(format #t "nfa-ref  call ~s ~s~%"  (car trans) state)
              (nfa-ref (car trans) state)
           )

          ((eq? A 'back-pos)
                (set! i (car trans))
                (set! op (cadr trans))
                (set! ind  (Rindex state))
                (set-car! op ind)
                (if (= i 0)
                   (begin
                       (set-Rindex state -1)
                       state
                    )
                   (begin
                      (set! i (- ind i))
                      (if (< i -1)
                            #f
                          (begin
                              (set-Rindex state i)
                              state
                          )
                       )
                   )
               )
            )
           ((eq? A 'inc-chk-pos)
                (set! op (car trans))
                (set! ind  (Rindex state))
                 (inc ind)
                   ;;;(format #t "inc-chk op ~s ind ~s~%" op ind)
                (if (<=  ind (car op) )
                    (begin
                      (set-Rindex state ind)
                      state
                     )
                  #f
                )
            ) 
          ((eq? A 'range)
              (range trans i state)
            )
          ((eq? A 'g-mode)
             ;;;;(format #t "g-mode called ~s~%" trans)
             (set-g-mode (car trans))
             state
            )
          ((eq? A 'FAIL)
              #f
            )
          (else
             (error 411 (cons A trans))
         )
    )
))

(define (range trans i state)
     (let* ((op (pop trans)))
    ;;;(format #t "range called ~s~%" trans)
             (cond ((eq? op :reset)
                            ;;;;(set-range-index state (car trans) 0) 
                            (set-car! (car trans) 0) 
                          state)
                   ((eq? op :ge)
                       ;;;(if (>=  (range-index  state (car trans)) (cadr trans))
                       (if (>=  (car (car trans)) (cadr trans))
                                   state
                                  #f
                       ) )
                    ((eq? op :le)
                       ;;;;(set! i (range-index state (car trans)))
                       (set! i (car (car trans)))
                       (inc i)
                       (if (<=  i (cadr trans))
                                (begin
                                  ;;;(set-range-index state (car trans)  i )
                                  (set-car! (car trans)  i )
                                   state
                                 )
                                  #f
                       ) )
                    (else
                       #f  )
             )
))

(define (regexp-string trans str len state)
   (let* ((i (Rindex state)) (cont #t) (ans 0))
     (while (and cont (pair? trans) (<= i len))
        (set! ans  (R-char=? state (car trans) (R-string-ref  i)))
        (if ans
           (begin
             (set! trans (cdr trans))
             (if (number? ans) 
                  (set! i ans)
                   (inc i)
             )
              (set-Rindex state i)  ;;;  for top end  \1
           )
             (set! cont #f)
        )
     )
   (if (and cont (null? trans))
     (begin
        (set-Rindex state i)
        state
     )
     #f
   )
))
       
(define (regexp-pure-string trans str len state)
  (R-cmp-string-main  (car trans) state)
)

                                  
;;;  subst  using only regexp and get-flag 
;;;   So this has  no changes from ver1.3
(define (regexp-subst str pat rstr . flag)
   (let* ((ind 0) (index 0) (res ())(from 0) (f 0) (subf #f))
     ;;;;(set! opt (get-flag flag))
     (set! index (apply regexp str pat :index flag))
     (if index
       (begin
          (if (not (string? rstr)) (set! subf rstr))
          (if (not *regexp-repeat*)(set! index (cons index ())))
          (while (pair? index)
             (set! ind (pop index))
             (push (substring str from (car ind)) res)
              (if subf
                  (set! rstr (funcall subf (substring str (car ind) (cdr ind))))
              )
             (push rstr res)
             (set! from (cdr ind))
          )
          (push (substring str from (string-length str)) res)
          (apply string-append (reverse res))
        )
        #f
    )
))
;;;;;
(define (regexp-group  str pat . flag)
   (set! flag (get-flag flag))
    (init-regexp str pat)
   (regexp-exec   (logior flag 8) )
)

(define (R-group-get state )
   ;;;(format #t "R-group state ~s ~%" state )
  (let* ((a 0) (res ()) (M *group-num*) (tmp ())(x 0)
         (gr (reverse (get-group state))) (v (make-vector (+ M 1) ())) (n 0))
   ;;;;(format #t "R-group ~s ~%" gr )
    (while (pair? gr)
       (set! a (pop gr)) 
  ;;;; (format #t "R-group a ~s  ~%" a)
          (set! n (cadar a))  
          (vector-set! v n (cons (cdr a)(vector-ref v n)))
   )
   (set! n M)
   (while (>  n 0)
        (set! a (vector-ref v n))
        (if (null? a)
            (push #f res)
            (begin
               (set! tmp ())
               (while (pair? a)
     ;;;(format #t "R-group  tmp a ~s ~%" a )
                  (set! x (pop a))
                  (push  (cons (pop a) x) tmp)
               )
              (push tmp res)
            )
         )
       (dec n)
    )
   ;;;;(format #t "R-group  res ~s ~%" res )
   (cons (state->index state) res)
))


(define (get-group-index n)
   (let* ((res *group-res*) )
     (when (string? n)
         (set! n (name->group n))
         (set! n (if n (car n) -1))
     )  
     (if (or (not res) (< n 0) (> n *group-num*))
        #f
        (begin
            (if *regexp-repeat* 
                (map (lambda(x) (list-ref x n)) res)
                (list-ref res n)
            )
        )
    )
))

(define (map1-safe  func lis)
   (if lis
       (map func lis)
       #f)
)

(define (get-group-string n)
   (let* ((res *group-res*) )
     (when (string? n)
         (set! n (name->group n))
         (set! n (if n (car n) -1))
     )
     (if (or (not res) (< n 0) (> n *group-num*))
        #f
        (begin
            (if *regexp-repeat*
                (if (= n 0)
                    ;;;;(map ind-substring (list-ref res n))
                    (map ind-substring   (map (lambda(x) (list-ref x n)) res) ) 
                    (map (lambda(x) (map1-safe  ind-substring (list-ref x n))) res)
               )
               (if (= n 0)
                  (ind-substring (list-ref res n))
                  (map ind-substring (list-ref res n))
               )      
            )
        )
    )
))


(define (last-term x)
  (do ((ans #f))
      ((not (pair? x)) ans)
      (set! ans (pop x))
  )
)
                                                                                              
(define (match-string  x )
    (str-group-ref x *regexp-state*)
)


(define (match-index  x )
    (index-group-ref x *regexp-state*)
)

              
(define (regexp-split str regex . opt)
  (let* ((x 0)(res ()) (s 0)(ans 0))
    (set! ans (apply regexp str regex :index :repeat opt))
    (if ans
         (begin
            (while (pair? ans)
               (set! x (pop ans))
               (push (substring str s (car x)) res)
               (set! s (cdr x))
            )
            (push (substring str s :end) res)
            (reverse res)
         )
       #f
     )
))

(init-R-debug)
             
(set! *auto-regexp* make-regexp)
(define ss:regexp regexp:regexp)    ;;;  ss:regexp =  tk:regexp overwrite!
(export '(regexp-group regexp-subst get-group-string get-group-index 
          regexp->nfa
          match-string match-index  make-regexp   regexp-split ))

(use-package "regexp" "ss")


                               
              

