;;;;  glob pattern matching 
;;;       specially for tcl program
;;;   04.5.4
;;;  ?,*, [...] ,  \c  (escape)
;;; use  (glob? "*\\*[a-z][0-9]"  "123*i7")  ->#t
;;;      (glob? "*\\*[a-z][0-9]"  "123-i7")  ->#f
;;;   Note :   in scheme string \  must be written as \\
;;;
(in-package "tclglob")

(define (glob? pat x)
   (glob-match (glob-pat pat) (string->list (tk-string x)))
)

(define (tk-conversion x)
 (let* ((res 0))
  (cond ((string? x) x)
        ((pair? x)
             (set! res (tk-conversion  (pop x)))
             (while (pair? x) 
                (set! res  (tk-string res " " (tk-conversion (pop x))))
             )
             res
           )
        (else (tk-string x))
   )
))

(define  (-glob pat x)
   (glob? pat (tk-conversion x))
)

(define (glob-match pat x)
  (let* ((a 0)(memo pat))
     (cond ( (null? pat)
               (null? x))
           ((null? x) (and (eqv? (length pat) 1) (eq? (car pat) '*)))
           (else
             (set! a (pop pat))
             (cond ((char? a)
                      (if (eq? a (pop x)) (glob-match pat x) #f)  )
                   ((eq? a '?)
                       (glob-match pat (cdr x))   )
                   ((eq? a '*)
                       (if (glob-match pat x) 
                            #t
                           (glob-match memo (cdr x))
                        )
                     )
                   (else
                     (if (glob-range? a (pop x))
                           (glob-match pat x) 
                           #f
                     )
                   )
              )
          )
     )
))
                    

;;;  ?,*, [...] ,  \c  (escape)
(define (glob-pat pat)
  (let* ((s (string->list pat)) (a 0)(res ()))
    (while (not (null? s))
      (set! a (pop s))
      (cond ((eq? a #\?) (push '?  res))
            ((eq? a #\*) (push '*  res))
            ((eq? a #\[) (set! a (get-glob-range s))
                         (push (car a) res)
                         (set! s (cdr a)))
           ((eq? a #\\)  (push (pop s) res) )
           (else  (push a res))
       )
     )
   (reverse res)
))

;;; caution!!!   not unicode
;;;
(define (get-glob-range s)
  (let* ((a 0)(b 0)(cont #t) (res ()) )
     (while (and  cont (not (null? s)))
          (set! a (pop s))
          (set! b (car s))
          (cond ((eq? b #\] )
                   (push a res)
                   (set! cont #f)
                   (pop s)
                  )
                ((eq? b #\-)
                   (pop s)
                   (push (list (char->integer a) (char->integer (pop s))) res) 
                   (if (eq? (car s) #\] )
                       (begin (set! cont #f) (pop s)) )
                   )
                (else
                   (push a res)
                   )
            )
       )
    (if cont (error 6543 "illegal glob-pattern"))
    (cons  res s)
))
  
(define (glob-range? bag x)
  (let* ((a 0) (c 0) (s 0)(e 0) (ans #f))
     (if (memq x bag)
           #t
      (begin
        (set! c (char->integer x))
        (while  (pair? bag)
          (set! a (pop bag))
          (if (pair? a )
            (begin
              (set! s (car a))
              (set! e (cadr a))
              (if (and (<= s c ) (<= c e))
                  (begin (set! bag ()) (set! ans #t)))
             )
           )
         )
        ans
      )
    )
))
                  
(export '(glob? -glob tk-conversion))
(use-package "tclglob" "ss")
              

