;;;;  vm-compiler.ss
;;;;  2007.1.8
;;;;  2007.8.28   Tail call version
;;;;   see tail-m.ss
;;;;
;;;;   2010.4.25   revised
;;;;   2010.5.9    NIL->MULTI

;;;(in-package "vmc")
;;; undef
(define UNDEF (nop))
(define Level 0)
(define *label* 10)   ;;; 0-9 are reserved
;;;  register
(define SCM UNDEF)
(define ENV ())
(define GENV ())
(define MULTI ())
(define BA  (list #f))     ;;;;  list of (before  . after)
(define Level-stack ())    ;;;;  for TAIL-JMP only
;
(define Com-stack ())

(define (Com-out x)
    (push x Com-stack)
)

(define (Label)
  (inc *label*)
  *label*
)

(define (Compiler-init)
   (set! Com-stack ())
   (set! SCM UNDEF)
   (set!  ENV ())
   (set! Level 0)
   (set! *label* 10)
   (set!  MULTI ())
)
;;;;
(define (Compile S )
   (Compiler-init)
   (Compile->list S)
   ;;;(Compile->vector)
   (reverse Com-stack)
)
;;;
(define (Compile->list S)
    (cond ((number? S) (Com-out (list 'IMM S)) (set! SCM S))
          ((symbol? S) (Symbol-Com S))
          ((pair? S)  (Combi-Com S #f))
          (else (Com-out (list 'IMM S)) (set! SCM S))
          ;;;(else (format #t "Not implemented~s~%" S)(error 9999 S))
    )
)

(define (Symbol-Com s)
   (let* ((a (assq s ENV)))
       (if a
            (begin
                 (Com-out (list 'LOADENV s))
                 (set! SCM (cdr a))
            )
            (begin
                  (Com-out (list 'LOAD s))
                  (set! a (assq s GENV))
                  (if a
                      (set! SCM (cdr a))
                      (if (symbol-bind? s)
                           (set! SCM s)
                           (set! SCM  UNDEF)
                      )
                 )
           )
       )
))

(define (symbol-local? s)
    (assq s ENV)
)

(define (user-global? s)
    (assq s GENV)
)

(define NOP 'NOP)
(define INLINE 'INLINE)

(define (ENV-special? s)
    (let* ((x (assq s GENV)))
      (if x 
          #f
         (if (assq s ENV)
             #f
             (special? s)
         )
      )
))
 
(define (is_special? s)
  ;;;(or (special? s)
  (or (ENV-special? s)
      (cond ((eq? s 'NOP) 'NOP)
            ((eq? s 'INLINE) 'INLINE)
            ((eq? s 'CALL/CC) 'CALL/CC)
            ((eq? s 'apply) 'apply)
            (else #f)
      )
  )
)

(define (is_sr2macro? f)
   (let* ((v (assq f ENV)))
      (if v
            #f
        (if (symbol? f)
          (begin
             ;; (set! v (symbol-value f))
             ;; (eq? (structure? v) 'sr2macro)
            #f   ;;;  Not support in this program
          )
         #f
        )
      )
))
              

(define (Combi-Com s tail-f)
  (let* ((res ()) (top (car s)) (para (cdr s)) )
     ;;;(Compile->list top)
     (let* ((x (is_special? top)))   ;;; not SCM
           (if x 
                (begin
                  ;;; (pop Com-stack)
                  (Special-Com x para tail-f)
                )
                (if (is_sr2macro? top)
                 (Macro-Com top para tail-f) 
                 (Func-Com top para tail-f) 
                )
           )
     )
))

(define (Func-Com top para tail-f)
   (let* ((narg 0))
     (Compile->list top)
     (Com-out (list 'FUNC tail-f))
     (set! para (reverse para))  ;; ARGS is normal order
     (while (pair? para)
         (Compile->list (pop para))     
         (Com-out (list 'PUSH))
         (inc narg)
     )
   (if tail-f
    (Com-out (list 'TAIL-JMP narg (- Level (car Level-stack))))
    (Com-out (list 'JSR narg))
   )
))    

(define (Macro-Com top para tail-f)
   (let* ((narg 0))
     (Compile->list top)
     (Com-out (list 'FUNC tail-f))
     (set! para (reverse para))  ;; ARGS is normal order
     (while (pair? para)
         (Com-out (list 'IMM (pop para)))
         (Com-out (list 'PUSH))
         (inc narg)
     )
   (if tail-f
    (Com-out (list 'TAIL-JMP narg (- Level (car Level-stack))))
    (Com-out (list 'JSR narg))
   )
))  

(define  (Apply-com para tail-f)
   (let* ((x 0))
     (Compile->list (pop para))
     (Com-out (list 'FUNC tail-f))
     (set! para (reverse para))
     (set! x (pop para))
     (Compile->list x)
     (Com-out (list 'SCM->ARGS))
     (while (pair? para)
         (Compile->list (pop para))
         (Com-out (list 'PUSH))
     )
   (if tail-f
    (Com-out (list 'TAIL-JMP () (- Level (car Level-stack))))
    (Com-out (list 'JSR ()))
   )
))

     
;;;
(define (Special-Com x para tail-f)
   (cond ((eq? x  'begin) (Begin-com para tail-f))
         ((eq? x 'lambda) (Lambda-com 'LAMBDA para))
         ((eq? x 'macro) (Lambda-com  'MACRO para))
         ((eq? x 'let*) (Let*-com para tail-f))
         ((eq? x 'letrec) (Letrec-com para tail-f))
         ((eq? x 'define) (Define-com para))
         ((eq? x 'apply) (Apply-com para tail-f))
         ((eq? x 'set!) (Set!-com para))
         ((eq? x 'CALL/CC) (Call/cc-com para))
         ((eq? x 'if) (If-com para tail-f))
         ((eq? x 'NOP) )
         ((eq? x 'INLINE) (INLINE-com para) )
         ((eq? x 'quote) (QUOTE-com para))
         (else (error 1234 (tk-string "Not supported:" x)))
  )
)

(define (INLINE-com para)
    (Com-out para)
 )


(define (Begin-com para tail-f)
   (if tail-f
        (Lambda-body-com para)
        (Begin-com-simple para)
   )
)

(define (Begin-com-simple para)
   (if (pair? para)
      (begin
        (Compile->list (car para))
        (Begin-com-simple (cdr para))
     )
   )
)

(define (Lambda-body-com para)
   (if (pair? para)
     (if (pair? (cdr para))
         (begin
           (Compile->list (car para))
           (Lambda-body-com (cdr para))
         )
        (tc-compile (car para))
     )
   )
)
;;;;  in lambda tail context
 (define   (tc-compile  f)
    (if (pair? f)
       (Combi-Com f #t)  ;;;  tail  form
       (Compile->list f)
    )
)



(define (Lambda-com x para)
  (let* ((env ENV) (obj 0)(com Com-stack))
      (set! Com-stack ())
      (inc Level)
      (push Level Level-stack)
     ;;;;(Com-out (list 'CLOSURE-ENV 0))   ;;;; ENV is set in VM
     ;;;(Com-out (list 'LAMBDA-BIND  (pop para)))
     (set! obj (pop para) )
     (Com-out (list 'LAMBDA-ARGS obj))
     (Lambda-bind obj 0)
     (Lambda-body-com para)
     (pop Level-stack)
     (Com-out (list 'RETURN))
      (dec Level)
      (set! ENV env)
      (set! obj (reverse Com-stack))
      (set! Com-stack com)
         (Com-out (list x obj))
     (set! SCM obj)
))

(define   (Lambda-bind args narg)
  (let* ((x 0))
    (if (pair?  args)
          (begin 
            (set! x (pop args))
            (Com-out (list 'LAMBDA-BIND x narg))
            (push (cons x UNDEF) ENV)
            (Lambda-bind args (+ narg 1))
         )
        (if (symbol? args)
            (begin
                 (Com-out (list 'LAMBDA-BIND (list  args) narg))
                 (push (cons args UNDEF) ENV)
            )
        )
    )
))    
          
     



(define (Let*-com para tail-f)
  (let* ((env ENV))
      (inc Level)
     (Com-out (list 'BLOCK Level))
     (Bind-para (pop para))
     (Begin-com para tail-f) 
     (Com-out (list 'BLOCKEND Level))
      (dec Level)
      (set! ENV env)
))

(define (Bind-para args)
  (let* ((x 0))
    (if (pair? args)
      (begin
         (set! x (pop args))
         (Bind-s (car x)(cadr x))
         (Bind-para args)
      )
    )
))

(define (Bind-s s val)
   (Compile->list val)
   (Com-out (list 'BIND s)) 
   (push (cons s SCM) ENV)
)

(define (Letrec-com para tail-f)
  (let* ((env ENV))
      (inc Level)
     (Com-out (list 'BLOCK Level))
     (Bind-para-rec (pop para))
     (Begin-com para tail-f)
     (Com-out (list 'BLOCKEND Level))
      (dec Level)
      (set! ENV env)
))

(define (Bind-para-rec args)
  (let* ((x 0)(memo args))
    ;;;(format #t "para-rec ~s~%" args)
    (while (pair? args)
         (set! x (pop args))
     ;;;  (format #t "para-rec x ~s~%" x)
         (Bind-s (car x) UNDEF)
    )
   (set! args memo)
   (while (pair? args)
         (set! x (pop args))
         (Compile->list (list 'set! (car x) (cadr x)))
    )
))

    
(define  (Define-com para)
  (let* ((x 0))
    (if (pair? para)
             (set! x (pop para))
             (error 555 para)
    )
    (if (not (symbol? x) ) (error 556 x))
             
    (cond ((eq? para ()) (Com-out (list 'IMM UNDEF))(set! SCM UNDEF))
         ((pair? (cdr para)) (error 555 para))
         (else  (Compile->list (car para)))
    )
    (if (= Level 0)
      (begin
         (Com-out (list 'DEFINE x))
         (push (cons x SCM) GENV)
      )
      (begin
         (Com-out (list 'BIND x))
         (push (cons x SCM) ENV)
      )
   )
))

(define  (Set!-com para)
   (let* ((x 0)(v 0))
    (if (pair? para)
             (set! x (pop para))
             (error 555 para)
    )
    (if (not (symbol? x) ) (error 556 x))
                                                                                                
    (cond ((eq? para ()) (error 555 para))
         ((pair? (cdr para)) (error 555 para))
         (else  (Compile->list (car para)))
    )
    (set! v (symbol-local? x))
    (if v
      (begin
         (Com-out (list 'SETENV  x))
         (Com-out (list 'IMM  UNDEF))
        ;;; (set-cdr! v SCM)   ;;;; this destroy VM env!!!
      )
     (begin
         (Com-out (list 'SET x))
         (Com-out (list 'IMM  UNDEF))
         (set! v  (user-global? x))
         (if v
              ;; (set-cdr! v SCM)   ;;; this destroy VM env!!!
            (nop)
            (begin
              (if (symbol-bind? x)
                    (warning 0  "gloval variable ~s is changed  to ~s~%"  x SCM) 
                    (warning 1  "Undefined gloval variable ~s is set  to ~s~%"  x SCM) 
              )
              (push (cons x SCM) GENV)
            )
        )    
      )
   )
))

(define (warning level format-str . para)
   (format #t "Warning Level ~s:" level)
   (apply format #t format-str para)
)


(define  (If-com para tail-f)
   (let* ((c 0)(v 0) (n (Label)))
      (if (not (pair? para))
             (error 559 para)
      )
      (set! c (pop para))
      (if (not (pair? para)) (error 559 para))
       (Compile->list c)
      (Com-out (list 'IF-FALSE n))
      (set! v (pop para))
      (if tail-f
        (tc-compile v)
        (Compile->list v)
      )
      (if (pair? para)
           (begin
             (set! c (Label))
             (Com-out (list 'JMP  c))
             (Com-out (list 'LABEL  n))
             (set! v (pop para))
             (if tail-f
               (tc-compile v)
               (Compile->list v)
              )
             (Com-out (list 'LABEL  c))
           )
           (Com-out (list 'LABEL  n))
      )
))

(define  (QUOTE-com para)
   (let* ((c 0))
      (if (not (pair? para))
             (error 559 para)
      )
      (set! c (pop para))
      (if  (pair? para) (error 559 para))
      (Com-out (list 'IMM  c))
))
     


(define (Call/cc-com para)
   (let* ((c 0))
      (if (not (pair? para))
             (error 559 para)
      )
      (set! c (pop para))
      (if (pair? para) (error 559 para))
      ;;;(Compile->list (list make-cc c ))
      ;;;(Compile->list (list 'let* (list (list 'proc c)) make-cc  ))
      (Compile->list (list let* (list (list 'proc c)) make-cc))
))


(define  make-cc  
      '(letrec ((CC #f) 
                     (CONT
                           (lambda () 
                                        (INLINE COPY-CC) ;;;continuation->SCM
                                        (INLINE SCM->ST)
                                        (if CC    ;; here SCM is changed
                                            (INLINE ST->SCM)
                                           (begin
                                             (INLINE ST->SCM)
                                             (INLINE SETENV CC) ;;;;set  CC
                                             (proc exit)
                                           )
                                         )
                                   ) )
                       (exit
                              (lambda values
                                      values
                                     (INLINE SCM->MULTI)
                                     CC
                                     (INLINE RESTORE-CC)   ;;; include 
                                  ) 
                               )
                 )
           (CONT)
        )
)


